#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 10 20:24:53 2018

@author: max
"""
from __future__ import division
import numpy as np
from scipy.special import jv
import sys
from math import sqrt
from Functions.CalcQuadratic import calc_quadratic


class DensityLimit:
    def __init__(self, core, nbi):
        sys.dont_write_bytecode = True
        self.r = core.r[:, 0]
        self.a = core.a
        self.eq44(core, nbi)
        pass

    def eq44(self, core, nbi):
        def av(X, p=self):
            numerator = np.sum(p.r * X * jv(0, 5.5 * p.r / p.a) * p.a / (len(p.r) - 1))
            denominator = np.sum(p.r * jv(0, 5.5 * p.r / p.a) * p.a / (len(p.r) - 1))
            return numerator / denominator

        ni = core.n.i[:, 0]
        Ti = core.T.i.J[:, 0]
        n0 = ni[0]
        n_av = av(ni)
        f = n0 / n_av

        chi = core.chi.bohm[:, 0]
        chi_hat = av(ni * chi) / n_av
        D = 0.5
        D_hat = av(ni * D) / n_av

        a = core.a
        g = ni / n0

        fz = 0.05
        Lz = core.Lz.t[:, 0]
        dLzdT = core.Lz.ddT.t[:, 0]

        sv_fus = core.sv.fus.dd[:, 0]
        dsv_fus_dT = core.sv.fus.d_dT.dd[:, 0]
        Ua = 0

        H_aux = nbi.beams.D1.dPdV.v1D.W

        H_ohm = 0
        dHdT = 3 / (2 * Ti) * (H_ohm - H_aux)
        dHdn = 0

        nn = core.n.n.tot[:, 0]
        sv_ion = core.sv.ion.st[:, 0]
        dsv_ion_dT = core.sv.ion.d_dT.st[:, 0]

        dSdn = nn * sv_ion
        dSdT = ni * nn * dsv_ion_dT

        if False:
            print 'sv_ion = ', sv_ion
            print 'nn = ', nn
            print 'ni = ', ni
            print 'Ti = ', Ti
            print 'a = ', a
            print 'dSdn = ', dSdn
            print 'D_hat = ', D_hat
            print 'dHdn = ', dHdn
            print 'Ua = ', Ua
            print 'sv_fus = ', sv_fus
            print 'fz = ', fz
            print 'Lz = ', Lz

        ya = 3 * av(Ti) * (av(dSdn) - (5.5 / a) ** 2 * D_hat) - av(dHdn + 2 * ni * (1 / 4 * Ua * sv_fus - fz * Lz))
        yb = 3 * av(ni) * (av(dSdn) - (5.5 / a) ** 2 * D_hat) + 3 * av(Ti) * av(dSdT) - av(
            dHdT + ni ** 2 * (1 / 4 * Ua * dsv_fus_dT + fz * (-dLzdT)))
        yc = 3 * av(ni) * av(dSdT)

        if False:
            print 'ya = ', ya
            print 'yb = ', yb
            print 'yc = ', yc
            print

        y1, y2 = calc_quadratic(ya, yb, yc)

        if False:
            print
            print 'y1 = ', y1
            print 'y2 = ', y2

        t1y1 = chi_hat * (5.5 / a) ** 2 * av(g) + 2 * y1 * (fz * av(g * Lz) - av(1 / 4 * Ua * g * sv_fus))
        t2y1 = 2 * av(g ** 2 * (1 / 4 * Ua * dsv_fus_dT + fz * (-dLzdT)))
        t3y1 = 4 * (av(-dHdT) - y1 * av(dHdn)) * av(g ** 2 * (1 / 4 * Ua * dsv_fus_dT + fz * (-dLzdT)))
        t4y1 = chi_hat * (5.5 / a) ** 2 * av(g) + 2 * y1 * (fz * av(g * Lz) - av(1 / 4 * Ua * g * sv_fus)) ** 2

        t1y2 = chi_hat * (5.5 / a) ** 2 * av(g) + 2 * y2 * (fz * av(g * Lz) - av(1 / 4 * Ua * g * sv_fus))
        t2y2 = 2 * av(g ** 2 * (1 / 4 * Ua * dsv_fus_dT + fz * (-dLzdT)))
        t3y2 = 4 * (av(-dHdT) - y2 * av(dHdn)) * av(g ** 2 * (1 / 4 * Ua * dsv_fus_dT + fz * (-dLzdT)))
        t4y2 = chi_hat * (5.5 / a) ** 2 * av(g) + 2 * y2 * (fz * av(g * Lz) - av(1 / 4 * Ua * g * sv_fus)) ** 2

        if False:
            print
            print 't1y1 = ', t1y1
            print 't2y1 = ', t2y1
            print 't3y1 = ', t3y1
            print 't4y1 = ', t4y1
            print 't1y2 = ', t1y2
            print 't2y2 = ', t2y2
            print 't3y2 = ', t3y2
            print 't4y2 = ', t4y2

        nlim1 = (1 / f) * (t1y1 / t2y1) * (1 + sqrt(1 + t3y1 / t4y1))
        nlim2 = (1 / f) * (t1y2 / t2y2) * (1 + sqrt(1 + t3y2 / t4y2))
        nlim3 = (1 / f) * (t1y1 / t2y1) * (1 - sqrt(1 + t3y1 / t4y1))
        nlim4 = (1 / f) * (t1y2 / t2y2) * (1 - sqrt(1 + t3y2 / t4y2))

        if False:
            print
            print 'nlim1 = ', nlim1
            print 'nlim2 = ', nlim2
            print 'nlim3 = ', nlim3
            print 'nlim4 = ', nlim4
            print

        nlim = min(i for i in [nlim1, nlim2, nlim3, nlim4] if i > 0)

        if n_av > nlim:
            print 'n_av = ', n_av
            print 'nlim = ', nlim
            print 'Disruption predicted: YES'
        else:
            print 'n_av = ', n_av
            print 'nlim = ', nlim
            print 'Disruption predicted: NO'
        print



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from __future__ import division
import sys
from GT3.IOL import IOL
from GT3.SOL import Sol
from GT3.ReadInFIle import ReadInfile
from GT3.ImpRadiation import ImpRad
from GT3.Core import Core
from GT3.BeamDeposition import BeamDeposition
from GT3.DensityLimit import DensityLimit
from GT3.Marfe import Marfe
from GT3.RadialTransport.radial_transport import RadialTransport

try:
    from GT3.Neutrals import Neutrals
except ImportError:
    pass

class gt3:

    def __init__(self, inputFile=None, preparedInput = None, mode="coreonly", **kwargs):
        sys.dont_write_bytecode = True
        # Create shotlabel as an attribute of plasma class
        if "iolFlag" in kwargs:
            iolFlag = kwargs['iolFlag']
        else:
            iolFlag = True
        if "neutFlag" in kwargs:
            neutFlag = kwargs['neutFlag']
        else:
            neutFlag = True
        if "verbose" in kwargs:
            verbose = kwargs['verbose']
        else:
            verbose = False
        if inputFile:
            self.inputFile = inputFile
        if preparedInput:
            self.inp = preparedInput
        else:
            self.inp = ReadInfile(self.inputFile)
        self.core = Core(self.inp)
        self.iolFlag = iolFlag
        self.neutFlag = neutFlag
        self.verbose = verbose
        self.beamPowerFracOverride = None
        self.ntrl_cpu_override = False

        try:
            import neutpy
            self.neutpyLoaded = True
        except ImportError:
            self.neutpyLoaded = False

        if mode == 'coreonly':
            pass


        if mode == 'coreandsol':
            self.sol = Sol(self.inp, self.core)
        elif mode == 'thermaliol':
            self.iol = IOL(self.inp, self.core)
        elif mode == 'fulliol':
            self.iol = IOL(self.inp, self.core)
            self.nbi = BeamDeposition(self.inp, self.core, self.iol, pwrFracOverride=self.beamPowerFracOverride)
        elif mode == 'imp':
            self.imp = ImpRad(core=self.core)
        elif mode == 'ntrls':
            self._run_neutpy()
        elif mode == 'ntrlsandiol':
            self.iol = IOL(self.inp, self.core)
            self.nbi = BeamDeposition(self.inp, self.core, self.iol, pwrFracOverride=self.beamPowerFracOverride)
            self._run_neutpy()
        elif mode == 'nbi':
            if self.iolFlag:
                self.iol = IOL(self.inp, self.core)
                self.nbi = BeamDeposition(self.inp, self.core, self.iol, pwrFracOverride=self.beamPowerFracOverride)
            else:
                self.nbi = BeamDeposition(self.inp, self.core, pwrFracOverride=self.beamPowerFracOverride)
        elif mode == 'marfe_denlim':
            if self.iolFlag:
                self.iol = IOL(self.inp, self.core)
                self.nbi = BeamDeposition(self.inp, self.core, self.iol, pwrFracOverride=self.beamPowerFracOverride)
            else:
                self.nbi = BeamDeposition(self.inp, self.core, pwrFracOverride=self.beamPowerFracOverride)
            self._run_neutpy()
            self.imp = ImpRad(core=self.core)
            self.dl = DensityLimit(self.core, self.nbi)
            self.mar = Marfe(core=self.core)
        elif mode == 'marfe':
            if self.iolFlag:
                self.iol = IOL(self.inp, self.core)
                self.nbi = BeamDeposition(self.inp, self.core, self.iol, pwrFracOverride=self.beamPowerFracOverride)
            else:
                self.nbi = BeamDeposition(self.inp, self.core,  pwrFracOverride=self.beamPowerFracOverride)
            self._run_neutpy()
            self.imp = ImpRad(core=self.core)
            self.mar = Marfe(core=self.core)
        elif mode == 'allthethings':
            if self.iolFlag:
                self.iol = IOL(self.inp, self.core)
                self.nbi = BeamDeposition(self.inp, self.core, self.iol,  pwrFracOverride=self.beamPowerFracOverride)
            else:
                self.nbi = BeamDeposition(self.inp, self.core,  pwrFracOverride=self.beamPowerFracOverride)
            self._run_neutpy()
            self.imp = ImpRad(core=self.core)
            self.dl = DensityLimit(self.core, self.nbi)
            self.mar = Marfe(self.inp, self.core, self.imp)
        elif mode == 'radialtrans':
            if self.iolFlag:
                self.iol = IOL(self.inp, self.core)
                self.nbi = BeamDeposition(self.inp, self.core, self.iol,  pwrFracOverride=self.beamPowerFracOverride)
            else:
                self.nbi = BeamDeposition(self.inp, self.core,  pwrFracOverride=self.beamPowerFracOverride)
            self.sol = Sol(self.inp, self.core)
            self._run_neutpy()
            self.imp = ImpRad(z=None, core=self.core)
            self.rtrans = RadialTransport(self.core, self.iol, self.nbi, self.iolFlag, self.neutFlag)

    def _run_neutpy(self, reRun=False):
        if self.neutpyLoaded:
            self.ntrl = Neutrals(self.inp, self.core, cpus=self.ntrl_cpu_override)
            if reRun:
                self.ntrl.reRun(cpus=self.ntrl_cpu_override)
        else:
            print "NeutPy is not loaded. Cannot run Neutrals calculation"


    def override_NBI_Pwrfrac(self, frac):
        if isinstance(frac, list):
            self.beamPowerFracOverride = frac
        else:
            print "Please provide the NBI power fraction override as a list"

    def run_SOL(self):
        self.sol = Sol(self.inp, self.core)
        return self

    def run_IOL(self):
        self.iol = IOL(self.inp, self.core)
        return self

    def run_NBI(self, reRun=False):
        try:
            self.iol
        except AttributeError:
            print ("IOL module not run. Running now...")
            self.run_IOL()
        if self.iolFlag:
            self.nbi = BeamDeposition(self.inp, self.core, self.iol, reRun=reRun,
                                      pwrFracOverride=self.beamPowerFracOverride)
        else:
            self.nbi = BeamDeposition(self.inp, self.core,  pwrFracOverride=self.beamPowerFracOverride)
        return self

    def run_impurities(self):
        self.imp = ImpRad(core=self.core)
        return self

    def run_neutrals(self, reRun=False):
        self._run_neutpy(reRun=reRun)
        return self

    def override_ntrl_cpus(self, num):
        self.ntrl_cpu_override = num
        return self

    def run_density_limit(self):
        if self.iolFlag:
            self.nbi = BeamDeposition(self.inp, self.core, self.iol,  pwrFracOverride=self.beamPowerFracOverride)
        else:
            self.nbi = BeamDeposition(self.inp, self.core,  pwrFracOverride=self.beamPowerFracOverride)
        self.dl = DensityLimit(self.core, self.nbi)
        return self

    def run_marf(self):
        self.mar = Marfe(self.inp, self.core)
        return self

    def run_radial_transport(self):
        try:
            self.iol
        except AttributeError:
            print ("IOL module not run. Running now...")
            self.run_IOL()
        try:
            self.nbi
        except AttributeError:
            print ("NBI module not run. Running now...")
            self.run_NBI()
        try:
            self.ntrl
        except AttributeError:
            print ("Neutrals module not run. Running now...")
            self.run_neutrals()

        try:
            self.imp
        except AttributeError:
            print ("Impurity radiation module not run. Running now...")
            self.imp = ImpRad(z=6, core=self.core)
        self.rtrans = RadialTransport(self.core, self.iol, self.nbi, self.iolFlag, self.neutFlag)
        return self

    def disable_IOL(self):
        self.iolFlag = False
        print ("Re-running Radial Transport without IOL")
        try:
            self.rtrans
        except:
            self.run_radial_transport()
        return self

    def disable_neutrals(self):
        self.neutFlag = False
        print ("Running Radial Transport without neutral particles")
        try:
            self.rtrans
        except:
            self.run_radial_transport()
        return self


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from GT3.Core.Functions.DrawCoreLine import draw_core_line
from scipy.interpolate import UnivariateSpline
from shapely.geometry import LineString
from math import pi, sqrt


def create_surf_area_interp(rho2psinorm, psinorm2rho, psi_data, sep_pts, R0_a, a, sep_val):
    rho_vals = np.linspace(0, sep_val, 50, endpoint=True)
    r_vals = rho_vals * a
    psi_vals = rho2psinorm(rho_vals)

    surf_area = np.zeros(rho_vals.shape)
    # skip the first value. First value is zero.
    for i, psi_val in enumerate(psi_vals[1:]):
        if psi_val < 1:
            try:
                # try to get the actual length of the fs trace
                fs_line, fs_pts = draw_core_line(psi_data.R, psi_data.Z, psi_data.psi_norm, psi_val, sep_pts)
                surf_area[i + 1] = fs_line.length * 2 * pi * fs_pts.axis[0]
            except:
                # if too close to the seperatrix, it may return None. In this case, estimate perimeter using Ramanujan
                # TODO: Improve this estimation. Also, see comments in draw_core_line
                rho_val = psinorm2rho(psi_val)
                a_el = rho_val * a  # a of the ellipse
                b_el = rho_val * a * 1.5  # b of the ellipse, assumed kappa of 1.5. This needs to be improved

                surf_area[i + 1] = pi * (3*(a_el+b_el) - sqrt((3*a_el + b_el)*(a_el+3*b_el)))

        else:
            surf_area[i+1] = LineString(sep_pts).length * 2*pi*R0_a

    # For some reason, psi_vals will sometimes not be monotonically increasing, especially near the magnetic axis.
    # This prevents UnivariateSpline from working. To prevent this, we're just going to delete any points that are
    # lower than the previous point, prior to doing the fit. As long as the psi data isn't too wonky, this should be
    # fine.
    psinormvals_mi = []  # psivals that are monotonically increasing
    rvals_mi = []  # rvals corresponding to monotonically increasing psi
    rhovals_mi = []  # rhovals corresponding to monotonically increasing psi
    surfarea_mi = []  # surf_area corresponding to monotonically increasing psi
    for i, psi_val in enumerate(psi_vals):
        if i == 0:
            rvals_mi.append(0)
            rhovals_mi.append(0)
            psinormvals_mi.append(0)
            surfarea_mi.append(0)
        elif psi_val > psinormvals_mi[-1]:
            rvals_mi.append(r_vals[i])
            rhovals_mi.append(rho_vals[i])
            psinormvals_mi.append(psi_vals[i])
            surfarea_mi.append(surf_area[i])

    rvals_mi = np.asarray(rvals_mi)
    rhovals_mi = np.asarray(rhovals_mi)
    psinormvals_mi = np.asarray(psinormvals_mi)
    surfarea_mi = np.asarray(surfarea_mi)

    r2sa = UnivariateSpline(rvals_mi, surfarea_mi, k=3, s=0)
    rho2sa = UnivariateSpline(rhovals_mi, surfarea_mi, k=3, s=0)
    psinorm2sa = UnivariateSpline(psinormvals_mi, surfarea_mi, k=3, s=0)
    # r2sa = UnivariateSpline(r_vals, surf_area, k=3, s=0)
    # rho2sa = UnivariateSpline(rho_vals, surf_area, k=3, s=0)
    # psinorm2sa = UnivariateSpline(psi_vals, surf_area, k=3, s=0)

    return r2sa, rho2sa, psinorm2sa



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from Core.Functions.DrawCounterLine import draw_contour_line
from contours.quad import QuadContourGenerator
from shapely.geometry import LineString
import numpy as np


def draw_core_line(R, Z, psi, psi_val, sep_pts):
    # create contour generator
    c = QuadContourGenerator.from_rectilinear(R[0], Z[:, 0], psi)

    # draw contours with psi_val
    contours = c.contour(psi_val)

    if len(contours) == 1:
        # then we're definitely dealing with a surface inside the seperatrix
        x, y = draw_contour_line(R, Z, psi, psi_val, 0)
    else:
        # we need to find which of the surfaces is inside the seperatrix
        for j, line in enumerate(contours):
            x, y = draw_contour_line(R, Z, psi, psi_val, j)

            if (np.amax(x) < np.amax(sep_pts[:, 0]) and
                np.amin(x) > np.amin(sep_pts[:, 0]) and
                np.amax(y) < np.amax(sep_pts[:, 1]) and
                np.amin(y) > np.amin(sep_pts[:, 1])):
                # then it's an internal flux surface
                break
    pts = np.column_stack((x, y))
    line = LineString(pts)
    out_pt = pts[np.argmax(pts, axis=0)[0]]
    in_pt = pts[np.argmin(pts, axis=0)[0]]
    top_pt = pts[np.argmax(pts, axis=0)[1]]
    bot_pt = pts[np.argmin(pts, axis=0)[1]]
    fs_axis = [(out_pt[0]+in_pt[0])/2, (out_pt[1]+in_pt[1])/2]
    return line, fs_axis


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from collections import namedtuple
from GT3.Core.Functions.DrawCoreLine import draw_core_line


def calc_kappa_elong(psi_data, sep_pts):
    # get kappa and elongation from the psi_norm=0.95 flux surface
    fs_line, fs_pts = draw_core_line(psi_data.R, psi_data.Z, psi_data.psi_norm, 0.95, sep_pts)
    elong_a = (fs_pts.top[1] - fs_pts.bot[1]) / (fs_pts.out[0] - fs_pts.inn[0])
    tri_a = (fs_pts.axis[0] - fs_pts.top[0]) / (fs_pts.out[0] - fs_pts.axis[0])

    # get kappa and elongation near the magnetic axis
    fs_line, fs_pts = draw_core_line(psi_data.R, psi_data.Z, psi_data.psi_norm, 0.05, sep_pts)

    # a small number isn't guaranteed to find a contour. If one is not found,
    # start near zero increase psi_norm until we find one
    if fs_line is None:
        for psi_val in np.linspace(0.05, 1, 100):
            fs_line, fs_pts = draw_core_line(psi_data.R, psi_data.Z, psi_data.psi_norm, psi_val, sep_pts)
            if fs_line is None:
                pass
            else:
                break

    elong_0 = (fs_pts.top[1] - fs_pts.bot[1]) / (fs_pts.out[0] - fs_pts.inn[0])
    tri_0 = 0

    kappa = namedtuple('kappa', 'axis sep')(elong_0, elong_a)
    tri = namedtuple('tri', 'axis sep')(tri_0, tri_a)
    return kappa, tri



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

class debugInp():
    def __init__(self):
        self.ebeam = 76.56296774
        self.pbeam = 4.88
        self.rtang = 1.715
        self.BT0 = 2.0
        self.nbeams_loc='/home/jonathan/Dropbox/GTEDGE/MyPrograms/GTEDGE/lib/beams/NBeamsMDS/NBeams/bin/Release/nbeams'



#!/usr/bin/python
"""
Created on Thu Dec 28 02:13:04 2017

@author: Jonathan
"""
###############################################################################
#
#    sensitivity study of IOL and neutrals
#
###############################################################################

import graphs as graphs
import matplotlib.pyplot as plt
import pickle
import numpy as np
from math import ceil, floor, log10
#164436_3720 good shot
shotid=164436

timeid=3720

runid='j1900'


###############################################################################
#
#   Utilities
#
##############################################################################

def yRangeFind(inList):
    x=np.concatenate(inList)
    flat=x.flatten()
    flat.sort()
    #return [flat[2]*.75,flat[-2]*1.25]
    a=long(flat[2]*.75)
    b=long(flat[-2]*1.25)
    try:
        aval=1*10**int(floor(log10(a)))
        if (ceil(log10(b))-(log10(b))) <=.25:
            return [0 if (aval == 1) else aval, .75*10**int(ceil(log10(b)))]
        elif (ceil(log10(b))-(log10(b))) <=.5:
            return [0 if (aval == 1) else aval, .5*10**int(ceil(log10(b)))]
        elif (ceil(log10(b))-(log10(b))) <=.75:
            return [0 if (aval == 1) else aval, .25*10**int(ceil(log10(b)))]
        else:
            return [0 if (aval == 1) else aval, .1*10**int(ceil(log10(b)))]
    # negative arg in log exception
    except:
        return [flat[2]*.75,flat[-2]*1.25]

###############################################################################
#
#    Chi Comparison
#
###############################################################################

def chiComp1(shot, shotnoIOL):
    prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j$ w/ IOL',
                r'$Q^{diff}_{j,r}=Q^{total}_j$ w/out IOL']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out IOL"

    adjustments = {}
    caption = ["Ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out IOL correction" % (str(shotid), str(timeid)),
               "Neutrals calculation included"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi1), (prettyID[1], shotnoIOL.rtrans.chi.chi.i.chi1)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi1[-30:], shotnoIOL.rtrans.chi.chi.i.chi1[-30:]]),
                         #yrange=[0., 15],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.2,
                         marginLeft=0.095,
                         capAdj=0.15,
                         yLabelAdj=-.25,
                         xLabelAdj=-.05,
                         size=(16, 12))


def chiComp2(shot, shotnoIOL):
    prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j$ w/ IOL',
                r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j$ w/out IOL']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out IOL correction"

    adjustments = {0: 0.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out IOL" % (str(shotid), str(timeid)),
               "and convective heat flow, neutrals calculation included"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi2), (prettyID[1], shotnoIOL.rtrans.chi.chi.i.chi2)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi2[-30:], shotnoIOL.rtrans.chi.chi.i.chi2[-30:]]),
                         #yrange=[0., 15],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.2,
                         marginLeft=0.095,
                         capAdj=0.15,
                         yLabelAdj=-.25,
                         xLabelAdj=-.05,
                         size=(16, 12))


def chiComp3(shot, shotnoIOL):
    prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j$ w/ IOL',
                r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j$ w/out IOL']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out IOL correction"

    adjustments = {0: 1.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out IOL" % (str(shotid), str(timeid)),
               "convective heat flow, and work done on plasma, neutrals calculation included"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi3), (prettyID[1], shotnoIOL.rtrans.chi.chi.i.chi3)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi3[-30:], shotnoIOL.rtrans.chi.chi.i.chi3[-30:]]),
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.2,
                         marginLeft=0.095,
                         capAdj=-0.15,
                         yLabelAdj=-.25,
                         xLabelAdj=-.05,
                         size=(16, 12))


def chiComp4(shot, shotnoIOL):
    #prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j-Q^{visc}_j$ w/ IOL',
    #            r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j-Q^{visc}_j$ w/out IOL']
    prettyID = [r'$\chi_{j,r}$ w/ IOL correction',
                r'$\chi_{j,r}$ w/out IOL correction']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out IOL correction"

    adjustments = {0: 1.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out IOL correction" % (str(shotid), str(timeid)),
               "convective heat flow,  work done on plasma and visc. heat"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor[:-1]), [(prettyID[0], shot.rtrans.chi.chi.i.chi4[:-1]), (prettyID[1], shotnoIOL.rtrans.chi.chi.i.chi4[:-1])],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi4[-30:], shotnoIOL.rtrans.chi.chi.i.chi4[-30:]]),
                         #yrange=[-5., 5],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.025,
                         marginBottom=.20,
                         marginLeft=0.095,
                         capAdj=-0.15,
                         yLabelAdj=-.25,
                         xLabelAdj=-.05,
                         size=(16, 12))


def chieComp(shot, shotnoIOL):
    prettyID = [r'$\chi_e$ w/ IOL',
                r'$\chi_e$ w/out IOL']

    title = r"Comparison of Electron Heat Conductivity w/ and w/out IOL correction"

    adjustments = {0: 1.}
    caption = ["Main electron heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out IOL correction" % (str(shotid), str(timeid)),
               "neutrals calculation included"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.e), (prettyID[1], shotnoIOL.rtrans.chi.chi.e)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.e[-30:], shotnoIOL.rtrans.chi.chi.e[-30:]]),
                         #yrange=[0., 50],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_e$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.02,
                         marginBottom=.25,
                         marginLeft=0.095,
                         capAdj=-.15,
                         xLabelAdj=-0.05,
                         yLabelAdj=-0.,
                         size=(16, 12))


###############################################################################
#
#   Deactivate neutrals
#
###############################################################################

def chiComp1neuts(shot, shotnoneuts):
    prettyID = [r'$Q^{diff}_{i}=Q^{total}_j$ w/ neutrals',
                r'$Q^{diff}_{i}=Q^{total}_j$ w/out neutrals']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out neutrals"

    adjustments = {0: -1.}
    caption = ["Ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out neutrals, IOL corrected" % (str(shotid), str(timeid))]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi1), (prettyID[1], shotnoneuts.rtrans.chi.chi.i.chi1)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi1[-30:], shotnoneuts.rtrans.chi.chi.i.chi1[-30:]]),
                         #yrange=[0, 15.],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.055,
                         #                    marginBottom=.15,
                         marginLeft=.1,
                         yLabelAdj=-.25,
                         size=(16, 12))


def chiComp2neuts(shot, shotnoneuts):
    prettyID = [r'$Q^{diff}_{i}=Q^{total}_j-Q^{conv}_j$ w/ neutrals',
                r'$Q^{diff}_{i}=Q^{total}_j-Q^{conv}_j$ w/out neutrals']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out neutrals"

    adjustments = {0: -.5}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%sw/ and w/out neutrals" % (str(shotid), str(timeid)),
               "IOL corrected, convective heat subtracted"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi2), (prettyID[1], shotnoneuts.rtrans.chi.chi.i.chi2)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi2[-30:], shotnoneuts.rtrans.chi.chi.i.chi2[-30:]]),
                         #yrange=[0, 15.],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.055,
                         marginBottom=.15,
                         marginLeft=0.095,
                         yLabelAdj=-.25,
                         size=(16, 12))


def chiComp3neuts(shot, shotnoneuts):
    prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j$ w/ neutrals',
                r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j$ w/out neutrals']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out neutrals"

    adjustments = {0: -1.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out neutrals" % (str(shotid), str(timeid)),
               "convective heat flow, and work done on plasma, IOL corrected"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi3), (prettyID[1], shotnoneuts.rtrans.chi.chi.chi3)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi3[-30:], shotnoneuts.rtranschi.chi.i.chi3[-30:]]),
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.15,
                         marginLeft=0.095,
                         yLabelAdj=-.25,
                         size=(16, 12))


def chiComp4neuts(shot, shotnoneuts):
    prettyID = [r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j-Q^{visc}_j$ w/ neutrals',
                r'$Q^{diff}_{j,r}=Q^{total}_j-Q^{conv}_j-Q^{heatin}_j-Q^{visc}_j$ w/out neutrals']

    title = r"Comparison of Ion Heat Conductivity w/ and w/out neutrals"

    adjustments = {0: -1.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out neutrals" % (str(shotid), str(timeid)),
               "convective heat flow,  work done on plasma and visc., IOL corrected"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.i.chi4), (prettyID[1], shotnoneuts.rtrans.chi.chi.i.chi4)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.i.chi4[-30:], shotnoneuts.rtrans.chi.chi.i.chi4[-30:]]),
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.2,
                         marginLeft=0.095,
                         capAdj=0.1,
                         xLabelAdj=-0.05,
                         yLabelAdj=-.25,
                         size=(16, 12))

def paperComp(shotIOL, shot):
    prettyID = [r'$\chi_{j,r}$ no corr',
                r'$\chi_{j,r}$ w/ IOL',
                r'$\chi_{j,r}$ w/ IOL, con',
                r'$\chi_{j,r}$ w/ IOL, conv, int']

    title = r"Comparison of Ion Heat Conductivity with various corrections"

    adjustments = {0: -1.}
    caption = ["Main ion heat conductivity in the edge for DIII-D shot %s.%s with no corrections (red)" % (str(shotid), str(timeid)),
               "corrected for IOL (blue), corrected for convective heat flux (green), and corrected for work done on plasma (purple)"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor[:-2]), [(prettyID[0], shot.rtrans.chi.chi.i.chi1[:-2]), (prettyID[1], shotIOL.rtrans.chi.chi.i.chi1[:-2]), (prettyID[2], shotIOL.rtrans.chi.chi.i.chi2[:-2]),
        (prettyID[3], shotIOL.rtrans.chi.chi.i.chi3[:-2])],
                         #yrange=yRangeFind([shot.rtrans.chi.chi.i.chi1[-30:], shotIOL.rtrans.chi.chi.i.chi3[-30:]]),
                         yrange=[-1.5, 3.5],
                         datalabels=[prettyID[0], prettyID[1], prettyID[2], prettyID[3]],
                         title=title,
                         ylabel=[r"$\chi_j$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.15,
                         marginLeft=0.095,
                         yLabelAdj=-.25,
                         size=(16, 12),
                         legend=True)



    prettyID = [r'$D_{drag,j}$']

    title = r"Calculation of $D$"

    adjustments = {0: -1.}
    caption = ["Main ion particle diffusion coefficient in the edge for DIII-D shot %s.%s with no corrections (red)" % (str(shotid), str(timeid))]

    graphs.prettyCompare(('rhor', shot.rtrans.nu_drag_D[:-2]), [(prettyID[0], shotIOL.rtrans.nu_drag_D[:-2])],
                         #yrange=yRangeFind([shot.rtrans.chi.chi.i.chi1[-30:], shotIOL.rtrans.chi.chi.i.chi3[-30:]]),
                         yrange=[-15, 35],
                         datalabels=[prettyID[0]],
                         title=title,
                         ylabel=[r"$D_{j}$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.065,
                         marginBottom=.15,
                         marginLeft=0.095,
                         yLabelAdj=-.25,
                         size=(16, 12),
                         legend=True)

def chieCompneuts(shot, shotnoneuts):
    prettyID = [r"$\chi_e$ w/ neutrals",
                r"$\chi_e$ w/out neutrals"]

    title = r"Comparison of Electron Heat Conductivity w/ and w/out neutrals"

    adjustments = {0: -1.}
    caption = ["Main electron heat conductivity in the edge for DIII-D shot %s.%s w/ and w/out neutrals" % (str(shotid), str(timeid)),
               "IOL corrected"]

    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.chi.chi.e), (prettyID[1], shotnoneuts.rtrans.chi.chi.e)],
                         yrange=yRangeFind([shot.rtrans.chi.chi.e[-30:], shotnoneuts.rtrans.chi.chi.e[-30:]]),
                         #yrange=[0, 50.],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\chi_e$", r'$\left[\frac{m^2}{s}\right]$'],
                         caption=caption,
                         adjust=adjustments,
                         xTickremove=None,
                         textSpace=.02,
                         marginBottom=.2,
                         marginLeft=0.095,
                         xLabelAdj=-0.025,
                         capAdj=0.15,
                         yLabelAdj=-.05,
                         size=(16, 12))

    #
    # prettyID = [r"Term 1",
    #             r"Term 2"]
    #
    # title = r"Comparison of Electron Heat Conductivity Terms"
    #
    # adjustments = {}
    # caption = ["Terms in electron heat conductivity in the edge for DIII-D shot %s.%s" % (str(shotid), str(timeid)),
    #            "IOL corrected. Term1 = 'diffusive' piece, Term2 = 'convective' piece"]
    #
    # term1 = [a / (b * shot.rtrans.xk * c) for a, b, c in zip(shot.rtrans.qHeate, shot.rtrans.xne, shot.rtrans.xte)]
    # term2 = [-1.5 * a / b for a, b in zip(shot.rtrans.gameltemp, shot.rtrans.xne)]
    #
    # graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], term1), (prettyID[1], term2)],
    #                      yrange=[-10, 50],
    #                      datalabels=[prettyID[0], prettyID[1]],
    #                      title=title,
    #                      ylabel=[r"$\frac{\chi_e}{L_{T_e}}$", r'$\left[\frac{m^2}{s}\right]$'],
    #                      caption=caption,
    #                      adjust=adjustments,
    #                      xTickremove=None,
    #                      textSpace=.02,
    #                      marginBottom=.2,
    #                      marginLeft=0.095,
    #                      xLabelAdj=-0.025,
    #                      capAdj=0.15,
    #                      yLabelAdj=-.05,
    #                      size=(16, 12))


def fluxComp(shot, shotnoIOL, shotnoneuts):
    prettyID = [r'$\Gamma_{r,j}$ w/ neutrals',
                r'$\Gamma_{r,j}$ w/out neutrals']

    caption = ["Ion radial particle flux for DIII-D shot %s.%s w/ and w/out neutrals" % (str(shotid), str(timeid))]
    title = r"$\Gamma_{r,j}$  w/ and w/out neutrals"
    adjustments = {}
    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.gamma_diff_D), (prettyID[1], shotnoneuts.rtrans.gamma_diff_D)],
                         #yrange=(yRangeFind([shot.rtrans.gamma_diff_D[-30:], shotnoneuts.rtrans.gamma_diff_D[-30:]])),
                         yrange=[1E17, 1E20],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\Gamma_{r,j}$", r"$\left[\frac{\#}{m^2 s}\right]$"],
                         caption=caption,
                         adjust=adjustments,
                         textSpace=.02,
                         yLabelAdj=-.8,
                         xLabelAdj=-0.025,
                         marginLeft=0.15,
                         marginBottom=0.2,
                         size=(16, 19))

    prettyID = [r'$\Gamma_{r,j}$ w/ IOL corr.',
                r'$\Gamma_{r,j}$ w/out IOL corr']

    caption = ["Ion radial particle flux for DIII-D shot %s.%s w/ and w/out IOL Correction" % (str(shotid), str(timeid))]
    title = r"$\Gamma_{r,j}$  w/ and w/out IOL corr"
    adjustments = {}
    graphs.prettyCompare(('rhor', shot.rtrans.rhor), [(prettyID[0], shot.rtrans.gamma_diff_D), (prettyID[1], shotnoIOL.rtrans.gamma_diff_D)],
                         #yrange=yRangeFind([shot.rtrans.gamma_diff_D[-30:], shotnoIOL.rtrans.gamma_diff_D[-30:]]),
                         yrange=[0.,4E20],
                         datalabels=[prettyID[0], prettyID[1]],
                         title=title,
                         ylabel=[r"$\Gamma_{r,j}$", r"$\left[\frac{\#}{m^2 s}\right]$"],
                         caption=caption,
                         adjust=adjustments,
                         textSpace=.02,
                         yLabelAdj=-.8,
                         xLabelAdj=-0.025,
                         marginLeft=0.15,
                         marginBottom=0.2,
                         size=(16, 19))


if __name__ == "__main__":

    #   IOL Sensitivity

    shotargs = {'shotid': shotid,
                'timeid': timeid,
                'runid': runid,
                'nbRun': True,
                'IOL': True,
                'quiet': False,
                'reNeut': False,
                'neutrals': True,
                'gt3Method': 'radialtrans',
                'debug' : True}
    try:
        shot = pickle.load(open("outputs/s%s.%s.dat" % (str(shotid), str(timeid)), "rb"))
    except:
        try:
            shot = GTEDGE3_cli.runGT3(shotargs)
        except BaseException as e:
            raise(e)
        try:
            pass
#            with open("outputs/s%s.%s.dat"  % (str(shotid), str(timeid)), "wb") as f:
#                pickle.dump(shot, f)
        except BaseException as e:
             raise(e)
        try:
            f.close()
        except:
            pass


    shotargs = {'shotid': shotid,
                'timeid': timeid,
                'runid': runid,
                'nbRun': True,
                'IOL': False,
                'quiet': True,
                'reNeut': False,
                'neutrals': True,
                'gt3Method': 'radialtrans',
                'debug' : False}

    try:
        shotnoIOL = pickle.load(open("outputs/s%s.%s.noIOL.dat" % (str(shotid), str(timeid)), "rb"))
    except:
        try:
            shotnoIOL = GTEDGE3_cli.runGT3(shotargs)
        except Exception as e:
            raise Exception(e)
        try:
            pass
#            with open("outputs/s%s.%s.noIOL.dat" % (str(shotid), str(timeid)), "wb") as f:
#                pickle.dump(shot, f)
        except Exception as e:
            print e
        try:
            f.close()
        except:
            pass

#    chiComp1(shot,shotnoIOL)
#    chiComp2(shot,shotnoIOL)
    # chiComp3(shot,shotnoIOL)
    #chiComp4(shot, shotnoIOL)

    #chieComp(shot, shotnoIOL)
    #    qComp1(shot,shotnoIOL)
    #    qComp2(shot,shotnoIOL)
    #    qComp3(shot,shotnoIOL)

    #   Neutrals sensitivity

    shotargs = {'shotid': shotid,
                'timeid': timeid,
                'runid': runid,
                'nbRun': True,
                'IOL': True,
                'quiet': True,
                'reNeut': False,
                'neutrals': False,
                'gt3Method': 'radialtrans',
                'debug' : False}
    try:
        shotnoneuts = pickle.load(open("outputs/s%s.%s.noneuts.dat"  % (str(shotid), str(timeid)), "rb"))
    except:
        try:
            shotnoneuts = GTEDGE3_cli.runGT3(shotargs)
        except Exception as e:
            raise Exception(e)
        try:
            pass
#            with open("outputs/s%s.%s.noneuts.dat" % (str(shotid), str(timeid)), "wb") as f:
#                pickle.dump(shot, f)
        except Exception as e:
            print e
        try:
            f.close()
        except:
            pass

#    chiComp1neuts(shot, shotnoneuts)
#    chiComp2neuts(shot, shotnoneuts)
    #    chiComp3neuts(shot,shotnoneuts)
    #chiComp4neuts(shot, shotnoneuts)



    #    qComp1neuts(shot,shotnoneuts)
    #    qComp2neuts(shot,shotnoneuts)
    #    qComp3neuts(shot,shotnoneuts)
    #    qComp4neuts(shot,shotnoIOL)
    #
    #qieComp1(shot, shotnoIOL)
    paperComp(shot, shotnoIOL)
    fluxComp(shot, shotnoIOL, shotnoneuts)

    ####################################################################################
    #
    #
    #   TODO: Electron chi
    #
    ####################################################################################
#    chieCompneuts(shot, shotnoneuts.rtrans.)
    plt.show(block=True)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.interpolate import UnivariateSpline, interp1d
import numpy as np
from math import exp
import matplotlib.pyplot as plt

def calc_gamma_diff_method(r, a, part_src_nbi_tot, part_src_nbi_lost, izn_rate, dVdrho, iol_adjusted=False, F_orb=None, neutFlag=True,
                           verbose=False):
    dVdr = UnivariateSpline(r, dVdrho(r / a) / a, k=2, s=0)
    dF_orb = UnivariateSpline(r, F_orb, k=3, s=0).derivative()
    izn_rateint = UnivariateSpline(r, izn_rate, k=2, s=0)
    part_src_nbi_totint = UnivariateSpline(r, part_src_nbi_tot, k=2, s=0)
    part_src_nbi_lostint = UnivariateSpline(r, part_src_nbi_lost, k=2, s=0)
    iolPeak = np.where(dF_orb(r) == dF_orb(r).max())

    def f(t, gamma, sion, snbi, snbi_loss, dFdr, iolFlag, peak):
        if neutFlag:
            #if t/a >= 0.95:
                #S = snbi(t) + sion(t)
            #else:
            S = snbi(t)
        else:
            S = snbi(t)
        # Physically, if the IOL peak has occured, everything radially outward should be dFdr = 0.0 since F(r)
        # should equal 0.5 until r=1.0.66
        dFdrval = dFdr(t)
        if t >= peak:
            dFdrval = 0.0
        if iolFlag:
            return S - snbi_loss(t) - gamma * (dFdrval + 1) / (t + 0.003)
        else:
            return S - gamma * (1 / (t + 0.003))

    from scipy.integrate import ode

    gamma = ode(f).set_integrator('vode', with_jacobian=False)
    gamma.set_initial_value(0., 0.).set_f_params(izn_rateint, part_src_nbi_totint, part_src_nbi_lostint, dF_orb, iol_adjusted, r[iolPeak])
    dt = a / len(r)
    x, y = [], []
    while gamma.successful() and gamma.t < a:
        x.append(gamma.t+dt)
        y.append(gamma.integrate(gamma.t+dt))
    #gamma = UnivariateSpline(x, y, k=3, s=0)
    gamma = interp1d(x, np.array([float(b) for b in y]), kind="linear", fill_value="extrapolate")

    if verbose:
        plot = plt.figure()
        fig1 = plot.add_subplot(311)
        fig2 = plot.add_subplot(312)
        fig3 = plot.add_subplot(313)

        fig1.scatter(r, izn_rateint(r), color="green")
        fig1.scatter(r, part_src_nbi_totint(r), color="red")
        fig1.scatter(r, part_src_nbi_lostint(r), color="black")
        fig1.legend([r"$S_{ion}$", r"$S_{nbi,tot}$", r"$S_{nbi,lost}$"])
        fig1.set_xlim(0.85 * a, a)

        fig2.scatter(r, dF_orb(r), color="red")
        fig2.set_xlim(0.85 * a, a)

        fig3.scatter(r, gamma(r))
        fig3.set_xlim(0.85 * a, a)
        plt.show()

        return fig1, fig2, fig3

    return gamma(r)

def calc_gamma_int_method(r, part_src_nbi_tot, part_src_nbi_lost, izn_rate, iol_adjusted=False, F_orb=None, neutFlag=True):
    # Piper Changes: Added cylindrical integral method as a separate function. This will be set to a separate variable in the main code.
    gamma = np.zeros(r.shape)

    if not neutFlag:
        izn_rate = [0.] * len(izn_rate)

    # Boundary condition at magnetic axis. Needs to be in units of ions/m^3.
    # Only has the second term, since it's the center value. Also uses delta_r of the next point to avoid indexing into a non-existant location.
    # If not adjusted for IOL, part_src_nbi_lost = 0 anyway, so no need for an IOL check.
    gamma[0] = (part_src_nbi_tot[0] - 2*part_src_nbi_lost[0] + izn_rate[0])*(r[1] - r[0])
    gamma[1] = gamma[0] + (part_src_nbi_tot[1] - 2*part_src_nbi_lost[1] + izn_rate[1])*(r[1] - r[0])

    # You'll  prolly want to change this since it uses a dreaded for loop.
    for n in range(2,len(r)): # Pretty sure len() is still valid for multidimensional arrays.
        # The 2*part_src_nbi_lost is the factor of 2 in the fast IOL.
        if iol_adjusted:
            # Imported the exp() function for the thermal IOL attenuation.
            gamma[n] = (r[n-1]/r[n]) * gamma[n-1] * exp(-2*(F_orb[n]-F_orb[n-1])) + (part_src_nbi_tot[n] - 2.*part_src_nbi_lost[n] + izn_rate[n])*(r[n] - r[n-1])
        else:
            gamma[n] = (r[n-1]/r[n]) * gamma[n-1] + (part_src_nbi_tot[n] + izn_rate[n])*(r[n] - r[n-1])

    return gamma


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np

def calc_torque(beam, fforb, index=False):
    """ Calculates torque from a neutral beam (or beam component)

    torque = F * r_tan = (P/v) * r_tan = (P/sqrt(2E/m)) * r_tan = P * sqrt(m/(2E)) * r_tan

    :param fforb:
    :param index:
    :param beam: beam object with attributes z, m, a, en, pwr, rtan
    :return: torque
    """
    if index is not False:

        power = beam.P.W[index]
        energy = beam.E.J[index]
        mass = beam.m
        rtan = beam.rtang[index]
        torque = power * np.sqrt(0.5 * mass / energy) * rtan * (1.0 - fforb)  # Piper Changes: Included fast ion losses.
        return torque
    else:
        power = beam.P.W
        energy = beam.E.J
        mass = beam.m
        rtan = beam.rtang

        torque = power * np.sqrt(0.5 * mass / energy) * rtan * (1.0-fforb)  # Piper Changes: Included fast ion losses.
        return torque


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import griddata, interp1d
from shapely.geometry import Point, LineString

def calc_psi(rho, pts, psi_data):
    # move along line between m_axis and obmp and define psi values corresponding to rho values
    rho_line = LineString([Point(pts.axis.mag), Point(pts.obmp)])
    init_coords = np.zeros((0, 2))
    for i, rhoval in enumerate(rho[:,0]):
        pt_coords = np.asarray(rho_line.interpolate(rhoval, normalized=True).coords)[0]
        init_coords = np.vstack((init_coords, pt_coords))

    psi_vals = griddata(np.column_stack((psi_data.R.flatten(), psi_data.Z.flatten())),
                             psi_data.psi.flatten(),
                             init_coords,
                             method='cubic')
    psi_norm_vals = griddata(np.column_stack((psi_data.R.flatten(), psi_data.Z.flatten())),
                             psi_data.psi_norm.flatten(),
                             init_coords,
                             method='cubic')

    psi = interp1d(rho[:,0], psi_vals)(rho)
    psi_norm = interp1d(rho[:, 0], psi_norm_vals)(rho)

    # If rho goes all the way to 1 (the usual case) then force the last psi_norm value to be 1.
    # Otherwise things break later.
    if rho[-1,0] == 1.0:
        psi_norm[-1] = 1.0

    return psi, psi_norm




#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import griddata,  interp1d
from shapely.geometry import Point, LineString


def calc_rho2psi_interp(pts, psi_data, sep_val):
    rho_vals = np.linspace(0, sep_val, 100)

    ptsRZ = np.zeros((len(rho_vals), 2))

    obmp_line = LineString([Point(pts.axis.mag), Point(pts.obmp)])
    for i, rho in enumerate(rho_vals):
        ptsRZ[i] = np.asarray(obmp_line.interpolate(rho, normalized=True).coords)

    psi_vals = griddata(np.column_stack((psi_data.R.flatten(), psi_data.Z.flatten())),
                        psi_data.psi.flatten(),
                        ptsRZ,
                        method='cubic')

    psinorm_vals = griddata(np.column_stack((psi_data.R.flatten(), psi_data.Z.flatten())),
                            psi_data.psi_norm.flatten(),
                            ptsRZ,
                            method='cubic')

    psinorm_vals[0] = 0
    psinorm_vals[-1] = sep_val

    # For some reason, psi_vals will sometimes not be monotonically increasing, especially near the magnetic axis.
    # This prevents UnivariateSpline from working. To prevent this, we're just going to delete any points that are
    # lower than the previous point, prior to doing the fit. As long as the psi data isn't too wonky, this should be
    # fine.
    psivals_mi = []  # psivals that are monotonically increasing
    psinormvals_mi = []  # psinormvals corresponding to monotonically increasing psi
    rhovals_mi = []  # rhovals corresponding to monotonically increasing psi
    for i, psi_val in enumerate(psi_vals):
        if i == 0:
            rhovals_mi.append(0)
            psivals_mi.append(psi_vals[0])  # this is probably supposed to be zero as well, but we'll leave it for now
            psinormvals_mi.append(0)
        elif psi_val > psivals_mi[-1]:
            rhovals_mi.append(rho_vals[i])
            psivals_mi.append(psi_vals[i])
            psinormvals_mi.append(psinorm_vals[i])

    rhovals_mi = np.asarray(rhovals_mi)
    psivals_mi = np.asarray(psivals_mi)
    psinormvals_mi = np.asarray(psinormvals_mi)

    rho2psi = interp1d(rhovals_mi, psivals_mi, fill_value='extrapolate')
    rho2psinorm = interp1d(rhovals_mi, psinormvals_mi, fill_value='extrapolate')
    psi2rho = interp1d(psivals_mi, rhovals_mi, fill_value='extrapolate')
    psi2psinorm = interp1d(psivals_mi, psinormvals_mi, fill_value='extrapolate')
    psinorm2rho = interp1d(psinormvals_mi, rhovals_mi, fill_value='extrapolate')
    psinorm2psi = interp1d(psinormvals_mi, psivals_mi, fill_value='extrapolate')

    # rho2psi = interp1d(rho_vals, psi_vals, fill_value='extrapolate')
    # psi2rho = interp1d(psi_vals, rho_vals, fill_value='extrapolate')
    return rho2psi, rho2psinorm, psi2rho, psi2psinorm, psinorm2rho, psinorm2psi



#!/usr/bin/python2.7

#########################################
#
#   GTEDGE V3
#  
#  Main CLI script
#
#########################################


from GT3.gt3 import gt3
from GT3.GT3Prep import gt3Prep

import sys, argparse, warnings, datetime, os



def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def argFunc(a):
    argDic={}
    try:
        argDic['shotid']=int(a.shotid)
    except:
        raise NameError("shotid is not formatted correctly")
    try:
        argDic['timeid']=int(a.timeid)
    except:
        raise NameError("timeid is not formatted correctly")
    try:
        argDic['runid']=str(a.runid)
    except:
        raise NameError("runid is not formatted corectly")
    argDic['nbRun']=a.nbRun
    argDic['IOL']=a.IOL
    argDic['quiet']=a.q
    argDic['gt3Method']=a.gt3Method
    argDic['reNeut']=a.reNeut
    
    return argDic

def customwarn(message, category, filename, lineno, file=None, line=None):
    sys.stderr.write(warnings.formatwarning(message, category, filename, lineno))

def runGT3(shotargs):
    ###########################################################################
    #
    #   GTEDGE 3 with full core data and interpretation
    #   Uses gt3 as backend for background plasma calculations
    #
    #   Data provided as 201-point CSV files with the following naming conv.:
    #
    #   Input/GT3Profs_shotid_timeid.csv                Profiles w/ Core
    #   Input/GT3Consts_shotid_timeid.csv               Constants
    #   Input/nudraginputs_shotid_timeid_2v2.csv        GTEDGE profiles
    #   Input/GT3NBIConsts_144977_3000                  NBI-Relevent Constants
    #
    ############################################################################
    shotid = shotargs['shotid']
    runid = shotargs['runid']
    timeid = shotargs['timeid']
    IOL = shotargs['IOL']
    nbRun = shotargs['nbRun']
    quiet = shotargs['quiet']
    gt3Method = shotargs['gt3Method']
    reNeut = shotargs['reNeut']
    neutrals = shotargs['neutrals']
    debug = shotargs['debug']

    errLog = open("GT3err.%s.%s.log" % (shotid, timeid), "w+")
    errLog.write("\n")
    errLog.write("Time: %s \n" % str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
    errLog.write("%s.%s.%s  nbRun=%s IOL=%s \n" % (str(shotid), str(timeid), str(runid), str(nbRun), str(IOL)))

    sys.stderr = errLog
    warnings.simplefilter('always', UserWarning)
    warnings.simplefilter('always', RuntimeWarning)
    warnings.showwarning = customwarn


    print "shotid=%s   runid=%s    timeid=%s   IOL Correction=%s" % (str(shotid),str(runid),str(timeid),str(IOL))
    maxfile='togt3_d3d_'+str(shotid)+'_'+str(timeid)
    gt3Prep(shotid, timeid, runid, maxfile, quiet = quiet, genFiles=False)


    if reNeut==True:
        try:
            os.remove(os.path.join('inputs','%s_%s','gt3_%s_%s_neut.dat') % (str(shotid),(timeid), str(shotid),(timeid)))
        except:
            pass

    myPlasma = gt3(shotlabel=maxfile, mode=gt3Method, iolFlag = IOL, neutFlag = neutrals)


    return myPlasma




###############################################################################
#
#
#    MAIN PROGRAM as CLI Interface
#
#	usage: GTEDGE3_cli.py [-h] [-nbRun [NBRUN]] [-IOL [IOL]] shotid timeid runid
#
#	positional arguments:
#	shotid          DIII-D shot id
#	timeid          DIII-D time id
#	runid           DIII-D run id
#
#	optional arguments:
#	-h, --help      show this help message and exit
#	-nbRun [NBRUN]  Run NBeams (default: True)
#	-IOL [IOL]      Correct for IOL (default: True)
#
#
###############################################################################   
    
if __name__== "__main__":


    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("shotid",help="DIII-D shot id")
    parser.add_argument("timeid",help="DIII-D time id")
    parser.add_argument("runid",help="DIII-D run id")
    parser.add_argument("-nbRun",type=str2bool,nargs='?',#const=True,
                        default=True,help="Run NBeams")
    parser.add_argument("-IOL",type=str2bool,nargs='?',#const=True,
                        default=True,help="Correct for IOL")
    parser.add_argument("-reNeut",type=str2bool,nargs='?',#const=True,
                        default=True,help="Rerun the neutrals")
    praser.add_argument("-neutrals", type=str2bool, nargs='?',
                        default=True, help="Calculates neutrals")
    parser.add_argument("-gt3Method",type=str2bool,nargs='?',#const="brndiolneuts",
                        default="brndiolneuts",help="What GT3 method to run")
    parser.add_argument("-debug", type=str2bool, nargs='?',
                        default="False", help="Produce Radial Transport debug profiles")
    parser.add_argument("-q",action='store_true',help="Disable informational popus")
    

    try:
        args=parser.parse_args()
    except SystemExit as err:
        if err.code==2:
            parser.print_help()
        sys.exit(0)
    run(argFunc(args))




#!/usr/bin/python

##########################################################
#
#    Graphical presentation of various analysis results
#    
# 
##########################################################

#from matplotlib import rc
import matplotlib.pyplot as plt
from math import ceil,log
import numpy as np
import pandas as pd
import matplotlib.patches as mpatches


#from matplotlib import rc
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
#rc('text', usetex=True)

                  
def expDelv(data):
        # Experimental DelV
    expDelV=[]
    expDelVfluid=[]
    delViol=[]
    for vtordhat,torvhat,yy1,yy2 in zip(data['vtordhat'].values(),
                                        data['torvhat'].values(),
                                        data['yy1'].values(),
                                        data['yy2'].values()):
        expDelV.append((vtordhat+yy1)-(torvhat+yy2))
        expDelVfluid.append(vtordhat-torvhat)
        delViol.append(yy1-yy2)
#    caption=r'$\Delta V$ for various calculations using experimental Deuterium data'
    multiScatPlot(data['rhor'],[expDelV,expDelVfluid,delViol,nuDragMIOL(4,data)[2],nuDragMIOL(5,data)[2]],title=r"$V_D-V_C$",
                  legend=[r'$V^{tot,meas}_D-V^{tot,meas}_C$',
                          r'$V^{fluid,meas}_D-V^{fluid,meas}_C$',
                          r'$V^{iol}_D-V^{iol}_C$',
                          r'$V^{tot,pert}_D-V^{tot,meas}_C$ w/out IOL cor.',
                          r'$V^{fluid,pert}_D-V^{fluid,meas}_C w/ IOL corr.$'],
                  yLabel=r"$V_D-V_C$ (m/s)",
                  colors=['blue','red','black','green','orange'],fontsize=20)
                  
    
def gammaPlot(data,gamma,gammahat):
    
#   Plot gamma vs. gammahat
    
    plot=plt.figure()
    fig1=plot.add_subplot(111)
    fig1.set_xlabel(r'$\rho$',fontsize=20)
    fig1.set_ylabel(r'$\Gamma_{ion}$',fontsize=20)
    fig1.set_title('Radial particle flux comparison, edge')
    gamma=fig1.scatter(data.rhor,gamma,marker='o',color='red')
    gammahat=fig1.scatter(data.rhor,gammahat,marker='o',color='blue')
    fig1.legend((gamma,gammahat),(r'$\Gamma_{r}$',r'$\hat{\Gamma}_{r}$'),scatterpoints=1,loc='upper left',ncol=1,fontsize=18)
    
def gammaPlotFull(data):
#   Plot gamma vs. gammahat
    gamma=data.gamma
    gammahat=data.gammahat
    
    plot=plt.figure()
    fig1=plot.add_subplot(111)
    fig1.set_xlabel(r'$\rho$',fontsize=20)
    fig1.set_ylabel(r'$\Gamma_{ion}$',fontsize=20)
    fig1.set_title('Radial particle flux comparison, full plasma')
    gamma=fig1.scatter(data.rhor,gamma,marker='o',color='red')
    gammahat=fig1.scatter(data.rhor,gammahat,marker='o',color='blue')
    fig1.legend((gamma,gammahat),(r'$\Gamma_{r}$',r'$\hat{\Gamma}_{r}$'),scatterpoints=1,loc='upper left',ncol=1,fontsize=18)    
    plt.show()
    
    
def diffCoPlot(data,ndrag):
    
    plot=plt.figure()
    fig1=plot.add_subplot(111)
    fig1.set_xlabel(r'$\rho$',fontsize=20)
    fig1.set_ylabel(r'$D$',fontsize=20)
    fig1.set_title('Diffusion coefficient')
    NF=fig1.scatter(data['rhor'],diffCalcs(data,ndrag)[0],marker='o',color='red')
    VP=fig1.scatter(data['rhor'],diffCalcs(data,ndrag)[1],marker='o',color='blue')
#    fig1.set_ybound(lower=-1.,upper=2.)
    fig1.legend((NF,VP),(r'$D_{NF}$',r'$D_{VP}$'),scatterpoints=1,loc='upper right',ncol=1,fontsize=18)


def nudragCheck(data,ndrag):
    
    mag=[]
    for n,vphi,nudrag in zip(data['xni'].values(),
                             data['torvhat'].values(),
                             ndrag):
    
        magtemp=n*data['xmas1']*vphi*nudrag
        mag.append(magtemp)
                             
    plot=plt.figure()
    fig1=plot.add_subplot(111)
    fig1.set_xlabel(r'$\rho$',fontsize=20)
    fig1.set_ylabel(r'$n_jm_j\hat{V}_{\phi}\nu_{drag}$',fontsize=20)
    fig1.set_title("Total momentum input density")
    fig1.scatter(data['rhor'],mag,marker='o',color='blue')
    
    plot2=plt.figure()
    fig2=plot2.add_subplot(111)
    fig2.set_xlabel(r'$\rho$',fontsize=20)
    fig2.set_ylabel(r'$\nu_{drag}$',fontsize=20)
    fig2.set_title(r'$\nu_{drag}$ from perturbation theory')
    fig2.scatter(data['rhor'],ndrag,marker='o',color='red')

#    fig1.legend((nudrag,magplot,nuc12),(r'$\nu_{d}$',r'$\nu$',r'$\nu_{12}$'),scatterpoints=1,loc='upper right',ncol=1,fontsize=18)
#    fig1.legend((nudrag,magplot),(r'$\nu_{d}$',r'$\nu$'),scatterpoints=1,loc='upper right',ncol=1,fontsize=18)

def chiDebug(data,funcs):
    
    funcs.dataGen.newScatPlot(data.rhor,data.qhatHeati,ylabel="Qhat")
    funcs.dataGen.newScatPlot(data.rhor,data.chihat.heatin,ylabel="qhat_in")
    funcs.dataGen.newScatPlot(data.rhor,data.chihat.heatvisc,ylabel="qhat_visc")
    funcs.dataGen.newScatPlot(data.rhor,data.chihat.gamconvi,ylabel="qhat_conv")

    funcs.dataGen.newScatPlot(data.rhor,data.qHeati,ylabel="Q")
    funcs.dataGen.newScatPlot(data.rhor,data.chi.heatin,ylabel="q_in")
    funcs.dataGen.newScatPlot(data.rhor,data.chi.heatvisc,ylabel="q_visc")
    funcs.dataGen.newScatPlot(data.rhor,data.chi.gamconvi,ylabel="q_conv")
    
def prettyCompare(xdata,ydata,yrange=[0.,1.],ylabel="No Label",datalabels=False,
                  title="No Title",caption=False,floatCaption=None,
                  logSpace=False,corePlot=False,plotType="line",
                  adjust=None,textSpace=0.,size=(12,9),xTickremove=None,
                  yTickremove=None, marginLeft=None,marginRight=None,
                  marginBottom=None,marginTop=None,capAdj=0.,
                  xLabelAdj=0.,yLabelAdj=0., xlabel=r"Normalized minor radius ($\rho$)",
                  toSN=False,legend=False):

    if legend:textSpace=0.
    prettyData=[xdata]
    for a in ydata:
        prettyData.append(a)
    pandasFrame=pd.DataFrame.from_items(prettyData)

#    These are the "Tableau 20" colors as RGB.    

#    tableau20 = [(31, 119, 180), (174, 199, 232), (255, 127, 14), (255, 187, 120),    
#                 (44, 160, 44), (152, 223, 138), (214, 39, 40), (255, 152, 150),    
#                 (148, 103, 189), (197, 176, 213), (140, 86, 75), (196, 156, 148),    
#                 (227, 119, 194), (247, 182, 210), (127, 127, 127), (199, 199, 199),    
#                 (188, 189, 34), (219, 219, 141), (23, 190, 207), (158, 218, 229)]  

    tableau20 = [(255,0,0), (0,0,255), (0,255,0), (255,0,255),    
                 (128,0,0), (128,0,128), (0,0,128), (0,128,128)]              
    markerList=["*","x","s","o"]
    for i in range(len(tableau20)):    
        r, g, b = tableau20[i]    
        tableau20[i] = (r / 255., g / 255., b / 255.)      

    fig=plt.figure(figsize=size)
    
#   Remove the plot frame lines. They are unnecessary chartjunk.    
    ax = plt.subplot(111)    
    ax.spines["top"].set_visible(False)    
    ax.spines["bottom"].set_visible(False)    
    ax.spines["right"].set_visible(False)    
    ax.spines["left"].set_visible(False)    

    plt.subplots_adjust(bottom=marginBottom,left=marginLeft,top=marginTop,right=marginRight)
#   Ensure that the axis ticks only show up on the bottom and left of the plot.    
#   Ticks on the right and top of the plot are generally unnecessary chartjunk.    
    ax.get_xaxis().tick_bottom()    
    ax.get_yaxis().tick_left()   
      
#   Limit the range of the plot to only where the data is.    
#   Avoid unnecessary whitespace.    
    plt.ylim(yrange[0],yrange[1])    
    if corePlot==True:
        plt.xlim(0,0.85+textSpace)     # core
        xrange=[0,.85]
    else:
        plt.xlim(.9,1+textSpace)    # edge
        xrange=[.9,1.]

#   Make sure your axis ticks are large enough to be easily read.    
#   You don't want your viewers squinting to read your plot. 
    if logSpace==False:
        if toSN!=False:
            if yTickremove!=None:
                plt.yticks(np.linspace(yrange[0],yrange[1],num=6)[:(-1*yTickremove)], [str(round(x,2)) for x in np.linspace(yrange[0]/(1.*10**toSN),yrange[1]/(1.*10**toSN),num=6)], fontsize=30)    
            else:
#                print np.linspace(yrange[0]/(10**toSN),yrange[1]/(1.*10**toSN),num=6)
                plt.yticks(np.linspace(yrange[0],yrange[1],num=6), [str(round(x,2)) for x in np.linspace(yrange[0]/(1.*10**toSN),yrange[1]/(1.*10**toSN),num=6)], fontsize=26)    
        elif yTickremove!=None:
            plt.yticks(np.linspace(yrange[0],yrange[1],num=6)[:(-1*yTickremove)], [str(round(x,2)) for x in np.linspace(yrange[0],yrange[1],num=6)], fontsize=30)
        else:
            plt.yticks(np.linspace(yrange[0],yrange[1],num=6), [str(round(x,2)) for x in np.linspace(yrange[0],yrange[1],num=6)], fontsize=20)
            
    else:
        if yTickremove!=None:
            plt.yticks([log(a) for a in np.logspace(yrange[0],yrange[1],num=6)][:(-1*yTickremove)],[str(round(x,2)) for x in [log(a) for a in np.logspace(yrange[0],yrange[1],num=6)]], fontsize=26)
        else:
            plt.yticks([log(a) for a in np.logspace(yrange[0],yrange[1],num=6)],[str(round(x,2)) for x in [log(a) for a in np.logspace(yrange[0],yrange[1],num=6)]], fontsize=26)
    
    plt.tick_params(axis='x',pad=18)



    if xTickremove!=None:
        plt.xticks(np.linspace(xrange[0],xrange[1],num=6)[:(-1*xTickremove)], [str(round(x,2)) for x in np.linspace(xrange[0],xrange[1],num=6)], fontsize=26)    
    else:
        plt.xticks(np.linspace(xrange[0],xrange[1],num=6), [str(round(x,2)) for x in np.linspace(xrange[0],xrange[1],num=6)], fontsize=26)    

        
#   Provide tick lines across the plot to help your viewers trace along    
#   the axis ticks. Make sure that the lines are light and small so they    
#   don't obscure the primary data lines. 
    if logSpace==False:        
        for y in np.linspace(yrange[0],yrange[1],num=6):    
            plt.plot(np.linspace(xrange[0],xrange[1],num=6), [y] * len(np.linspace(xrange[0],xrange[1],num=6)), "--", lw=0.5, color="black", alpha=0.3)     
    else:
        for y in [log(a) for a in np.logspace(yrange[0],yrange[1],num=6)]:    
            plt.plot([log(a) for a in np.logspace(yrange[0],yrange[1],num=6)], [y] * len([log(a) for a in np.logspace(yrange[0],yrange[1],num=6)]), "--", lw=0.5, color="black", alpha=0.3)     
        
#   Remove the tick marks; they are unnecessary with the tick lines we just plotted.    
    plt.tick_params(axis="both", which="both", bottom=False, top=False,
                    labelbottom=True, left=False, right=False, labelleft=True)
    ystep=(yrange[1]-yrange[0])/25.
    if legend:
        plotList=[]
        patchList=[]
        for rank,column in enumerate(datalabels):
    #       Build the plots for plt.legend()
            x,=plt.plot(pandasFrame.rhor.values,
                         pandasFrame[column.replace("\n", " ")].values,
                        lw=2, color=tableau20[rank],marker=markerList[rank],markersize=10)
            plotList.append(x)
        plt.legend(plotList, datalabels, loc=2, fontsize=20,framealpha=1.)
    else:
        for rank, column in enumerate(datalabels):

    #       Plot each line separately with its own color, using the Tableau 20
    #       color set in order.
            if plotType=="line":
                plt.plot(pandasFrame.rhor.values,
                         pandasFrame[column.replace("\n", " ")].values,
                        lw=2.5, color=tableau20[rank])
                y_pos=pandasFrame[column.replace("\n", " ")].values[-1]
            elif plotType=="scat":
                plt.scatter(pandasFrame.rhor.values,
                        pandasFrame[column.replace("\n", " ")].values,
                        color=tableau20[rank])
                y_pos=pandasFrame[column.replace("\n", " ")].values[-1]
                if legend:
                    ax.legend(column, datalabels)
            else:
                raise ("No line type requested")
            try:
                if rank in adjust.keys():
                    y_pos += adjust[rank]*ystep
            except:
                pass

        # Again, make sure that all labels are large enough to be easily read
        # by the viewer.
            #plt.text(xrange[1]+0.0025, y_pos, column, fontsize=16, color=tableau20[rank]) # Default
            plt.text(xrange[1] + 0.0025, y_pos, column, fontsize=26, color=tableau20[rank])

    # matplotlib's title() call centers the title on the plot, but not the graph,    
    # so I used the text() call to customize where the title goes.    
      
    # Make the title big enough so it spans the entire plot, but don't make it    
    # so big that it requires two lines to show.    
      
    # Note that if the title is descriptive enough, it is unnecessary to include    
    # axis labels; they are self-evident, in this plot's case.

#    plt.text((xrange[1]-xrange[0])/2.+xrange[0]+0.015, (yrange[1]-yrange[0])/20.+yrange[1], title, fontsize=26, ha="center")      
    plt.text((xrange[1]-xrange[0])/2.+xrange[0]+.5*textSpace, (yrange[1]-yrange[0])/20.+yrange[1], title, fontsize=30, ha="center")

    if caption!=False:
        for num,cap in enumerate(caption):
            plt.text(xrange[0],yrange[0]-(yrange[1]-yrange[0])/5.+capAdj*(yrange[1]-yrange[0])/10.-(num)*(yrange[1]-yrange[0])/22.5, cap, fontsize=16,ha="left")
    if floatCaption!=None:
        plt.text(floatCaption[1][0],floatCaption[1][1],floatCaption[0],fontsize=22,ha="left",va="top")
        
    if not legend:
        for num,label in enumerate(ylabel):
            plt.text(xrange[0]-(xrange[1]-xrange[0])/10.+yLabelAdj*(xrange[1]-xrange[0])/10.,(yrange[1]-yrange[0])/2.+yrange[0]-num*(yrange[1]-yrange[0])/10.,label,fontsize=36,ha="center") # default


    #plt.text((xrange[1] - xrange[0]) / 2. + xrange[0],yrange[0] - (yrange[1] - yrange[0]) / 9. + xLabelAdj * (yrange[1] - yrange[0] / 10.), xlabel, fontsize=16, ha="center") #Default
    plt.text((xrange[1] - xrange[0]) / 2. + xrange[0],yrange[0] - (yrange[1] - yrange[0]) / 9. + xLabelAdj * (yrange[1] - yrange[0] / 10.), xlabel, fontsize=20, ha="center")


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.constants import e
from GT3.IOL.Functions.CalcVSep import calc_vsep
from scipy.special import gammaincc



def calc_iol_maxwellian(z, m, param, thetapts, Tprofile, coslist, numcos):
    """Calculates eps_min for IOL of species treated with a truncated maxwellian."""

    v_sep, v_sep_min = calc_vsep(z, m, param)

    # Define the Maxwellian for every point in the plasma based on its temperature

    T_matrix = np.zeros(v_sep_min.shape)
    for indx, row in enumerate(T_matrix):
        T_matrix[indx, :] = Tprofile[indx]

    # Create the launch angle matrix (used in the calculation of M_orb)
    zeta_matrix = np.zeros(v_sep_min.shape)
    for indx, column in enumerate(zeta_matrix.T):
        zeta_matrix[:, indx] = coslist[indx]

    eps_min = m * v_sep_min**2 / (2.*T_matrix*1E3*e)

    # F_orb calculation
    # note: the use of gammaincc renders the denominator in Dr. Stacey's equations obsolete.
    integrand_f = gammaincc(3./2., eps_min)
    F_orb_1D = np.sum(integrand_f, axis=1)*(2./numcos)/2.
    F_orb_1D = np.nan_to_num(F_orb_1D)
    F_orb = np.repeat(F_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # M_orb calculation
    integrand_m = zeta_matrix*gammaincc(2, eps_min)
    M_orb_1D = np.sum(integrand_m, axis=1)*(2./numcos)/2.
    M_orb_1D = np.nan_to_num(M_orb_1D)
    M_orb = np.repeat(M_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # E_orb calculation
    integrand_e = gammaincc(5./2, eps_min)
    E_orb_1D = np.sum(integrand_e, axis=1)*(2./numcos)/2.
    E_orb_1D = np.nan_to_num(E_orb_1D)
    E_orb = np.repeat(E_orb_1D.reshape(-1, 1), thetapts, axis=1)

    return F_orb, M_orb, E_orb



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Point, MultiPoint, LineString

def find_xpt_mag_axis(core, R, Z, psi):
    # type: (GT3.core, np.array, np.array, np.array) -> (np.array, np.array, np.array)
    """

    :type core: .core
    """
    xpt_u = None
    xpt_l = None
    m_axis = None
    # find x-point location
    dpsidR = np.gradient(psi, R[0, :], axis=1)
    dpsidZ = np.gradient(psi, Z[:, 0], axis=0)
    d2psidR2 = np.gradient(dpsidR, R[0, :], axis=1)
    d2psidZ2 = np.gradient(dpsidZ, Z[:, 0], axis=0)

    # find line(s) where dpsidR=0
    csR = plt.contour(R, Z, dpsidR, [0])
    csZ = plt.contour(R, Z, dpsidZ, [0])

    dpsidR_0 = csR.collections[0].get_paths()
    # dpsidR_0 = cntr.contour(R, Z, dpsidR).trace(0.0)

    # find line(s) where dpsidZ=0
    # dpsidZ_0 = cntr.contour(R, Z, dpsidZ).trace(0.0)
    dpsidZ_0 = csZ.collections[0].get_paths()
    for i, path1 in enumerate(dpsidR_0):
        for j, path2 in enumerate(dpsidZ_0):
            try:
                # find intersection points between curves for dpsidR=0 and dpsidZ=0
                ints = LineString(path1.vertices).intersection(LineString(path2.vertices))
                # if there is only one intersection ('Point'), then we're probably not
                # dealing with irrelevant noise in psi
                if isinstance(ints, Point):
                    # check if local maximum or minimum
                    # If the point is within the walls and less than 0.5m from the vessel centroid, this is
                    # very likely the magnetic axis
                    if core.wall_line.convex_hull.contains(ints) and core.wall_line.centroid.distance(ints) < 0.5:
                        # we've found the magnetic axis
                        m_axis = np.array([ints.x, ints.y])
                    # If the point is not inside the walls, it's a magnet
                    elif not core.wall_line.convex_hull.contains(ints):
                        # we've found a magnet. Do nothing.
                        pass
                    elif core.wall_line.convex_hull.contains(ints) and ints.y < 0:
                        # The point is within the walls and not near the magnetic axis. This is likely the
                        # lower x-point.
                        # TODO: Provides lower x-point here, but upper x-point can be built out
                        xpt = np.array([ints.x, ints.y])
                        xpt_l = xpt
                    elif core.wall_line.convex_hull.contains(ints) and ints.y > 0:
                        # This is an upper x-point. This functionality needs to be built in but is here for
                        # later
                        xpt_u = np.array([ints.x, ints.y])

                    # uncomment this line when debugging
                    # print list(ints.coords), d2psidR2(ints.x, ints.y), d2psidZ2(ints.x, ints.y)

                # If multiple points are found, the flux surfaces may have come back around onto each other.
                if isinstance(ints, MultiPoint):
                    for point in ints:
                        # check if local maximum or minimum
                        # If the point is within the walls and less than 0.5m from the vessel centroid, this is
                        # very likely the magnetic axis
                        if core.wall_line.convex_hull.contains(point) and core.wall_line.centroid.distance(
                                point) < 0.5:
                            # we've found the magnetic axis
                            m_axis = np.array([point.x, point.y])
                        # If the point is not inside the walls, it's a magnet
                        elif not core.wall_line.convex_hull.contains(point):
                            # we've found a magnet. Do nothing.
                            pass
                        elif core.wall_line.convex_hull.contains(point) and point.y < 0:
                            # The point is within the walls and not near the magnetic axis. This is likely the
                            # lower x-point.
                            # TODO: Provides lower x-point here, but upper x-point can be built out
                            xpt = np.array([point.x, point.y])
                            xpt_l = xpt
                        elif core.wall_line.convex_hull.contains(point) and point.y > 0:
                            # This is an upper x-point. This functionality needs to be built in but is here for
                            # later
                            xpt_u = np.array([point.x, point.y])

            except:
                pass
    return xpt_l, xpt_u, m_axis



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from math import sqrt

def calc_quadratic(a, b, c):
    y1 = -b * (1 + sqrt(1 - 4 * (a * c / b ** 2))) / (2 * a)
    y2 = -b * (1 - sqrt(1 - 4 * (a * c / b ** 2))) / (2 * a)
    return y1, y2


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.constants import e


def calc_vsep(z, m, p):
    """Calculates V_sep"""
    a = (np.abs(p.B1 / p.B0) * p.f0 / p.f1 * p.zeta0) ** 2 - 1 + (1 - p.zeta0 ** 2) * np.abs(p.B1 / p.B0)
    b = 2 * z * e * (p.psi0 - p.psi1) / (p.R1 * m * p.f1) * np.abs(p.B1 / p.B0) * p.f0 / p.f1 * p.zeta0
    c = (z * e * (p.psi0 - p.psi1) / (p.R1 * m * p.f1)) ** 2 - 2 * z * e * (p.phi0 - p.phi1) / m

    v_sep_1 = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)
    v_sep_2 = (-b - np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    v_sep = np.zeros(p.r0.shape)
    v_sep = np.where(
        np.logical_and(
            np.logical_or(v_sep_1 <= 0, np.isnan(v_sep_1)),
            np.logical_or(v_sep_2 <= 0, np.isnan(v_sep_2))
        ),
        np.nan, v_sep)
    v_sep = np.where(
        np.logical_or(
            np.logical_and(v_sep_1 > 0, np.logical_or(v_sep_2 <= 0, np.isnan(v_sep_2))),
            np.logical_and(np.logical_and(v_sep_1 > 0, v_sep_2 > 0), v_sep_1 < v_sep_2)
        )
        , v_sep_1, v_sep)
    v_sep = np.where(
        np.logical_or(
            np.logical_and(v_sep_2 > 0, np.logical_or(v_sep_1 <= 0, np.isnan(v_sep_1))),
            np.logical_and(np.logical_and(v_sep_1 > 0, v_sep_2 > 0), v_sep_2 <= v_sep_1)
        )
        , v_sep_2, v_sep)

    v_sep_min = np.nanmin(np.nanmin(v_sep, axis=0), axis=2).T
    v_sep_min[-1] = 0
    return v_sep, v_sep_min



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import interp1d
from GT3.IOL.Functions.CalcVSep import calc_vsep


def calc_iol_beams(z, m, param, thetapts, v_mono, zeta_beam, coslist):
    """calculates IOL for a monoenergetic species with a single known launch angle (i.e. beam ions)
    :param z: The Z value of the beam ions
    :type z: float
    :param m: The mass of the beam ions in kg
    :type m: float
    :param param: The IOL parameters
    :param thetapts: The number of theta points
    :type thetapts: int
    :param v_mono: The particle velocity in m/s
    :type v_mono: float
    :param zeta_beam: The zeta value
    :param coslist:
    :return:
    """

    v_sep, v_sep_min = calc_vsep(z, m, param)

    # Obtain v_sep_min(zeta_beam) for each rho value
    v_sep_min_zeta = np.zeros(len(v_sep_min))

    for i,v in enumerate(v_sep_min):
        # create interpolation function of v_sep_min(rho) vs zeta
        v_sep_min_zeta_interp = interp1d(coslist, v, fill_value='extrapolate')
        # get v_sep_min(zeta_beam)
        v_sep_min_zeta[i] = v_sep_min_zeta_interp(zeta_beam)

    # F_orb calculation
    F_orb_1D = np.heaviside(v_mono - v_sep_min_zeta, 1)
    F_orb_1D = np.nan_to_num(F_orb_1D)
    F_orb = np.repeat(F_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # M_orb calculation
    M_orb_1D = zeta_beam * np.heaviside(v_mono - v_sep_min_zeta, 1)
    M_orb_1D = np.nan_to_num(M_orb_1D)
    M_orb = np.repeat(M_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # E_orb calculation
    # E_orb is mathematically identical to F_orb for monoenergetic, monodirectional species
    E_orb = F_orb

    return F_orb, M_orb, E_orb



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.interpolate import UnivariateSpline
from math import sqrt, pi, exp
from GT3.RadialTransport.Functions.CalcCoulLog import calc_coul_log
from scipy import constants
import numpy as np
import matplotlib.pyplot as plt

eps_0 = constants.epsilon_0
e = constants.elementary_charge
m_e = constants.electron_mass
m_d = constants.physical_constants['deuteron mass'][0]
m_c = 12 / constants.N_A / 1E3  # in kg


def calc_Qe_diff_method(r, a, en_src_nbi_e, cool_rate, Qie):
    en_src_nbi_eint = UnivariateSpline(r, en_src_nbi_e, k=3, s=0)
    cool_rateint = UnivariateSpline(r, cool_rate, k=3, s=0)
    Qie_int = UnivariateSpline(r, Qie, k=3, s=0)

    def f(t, flux, Qie, Q_e_nbi, cool_rate):
        S = Q_e_nbi(t) - Qie(t) - cool_rate(t)
        return S - flux * (1 / (t + 0.003))

    from scipy.integrate import ode

    flux = ode(f).set_integrator('vode', with_jacobian=False)
    flux.set_initial_value(0., 0.).set_f_params(Qie_int, en_src_nbi_eint, cool_rateint)

    dt = a / len(r)
    x, y = [], []
    while flux.successful() and flux.t < a:
        x.append(flux.t + dt)
        y.append(flux.integrate(flux.t + dt))
    flux = UnivariateSpline(x, y, k=3, s=0)

    # print "Total volume in Qi_diff calc: " + str(UnivariateSpline(r, dVdr(r), k=3, s=0).integral(0., a))
    # print "Total nbi ion energy: " + str(UnivariateSpline(r, (en_src_nbi_keptint(r) + en_src_nbi_lostint(r)) * dVdr(r), k=3, s=0).integral(0., 1.)/(1E6))+" MW"
    return flux(r)

def calc_Qe_int_method(r, n, T, en_src_nbi_e_tot, cool_rate):  # Piper Changes: Same as Qi changes.

    Qe = np.zeros(r.shape)
    qie = calc_qie(n, T, ion_species='D')

    Qe[0] = (en_src_nbi_e_tot[0] - qie[0] - cool_rate[0]) * (r[1] - r[0])
    Qe[1] = Qe[0] + (en_src_nbi_e_tot[1] - qie[1] - cool_rate[1]) * (r[1] - r[0])

    # Integral cylindrical form of the energy balance equation.
    # Identical in form to the continuity equation, but different source term.
    for n in range(2, len(r)):
        Qe[n] = (r[n - 1] / r[n]) * Qe[n - 1] + (en_src_nbi_e_tot[n] - qie[n] - cool_rate[n]) * (r[n] - r[n - 1])

    return Qe


def calc_Qi_diff_method(r, a, cxcool, Qie, en_src_nbi_tot, en_src_nbi_lost, dVdrho, iol_adjusted=False, E_orb=None, verbose=False):
    dE_orb = UnivariateSpline(r, E_orb, k=3, s=0).derivative()
    en_src_nbi_totint = UnivariateSpline(r, en_src_nbi_tot, k=3, s=0)
    en_src_nbi_lostint = UnivariateSpline(r, en_src_nbi_lost, k=3, s=0)
    Qie_int = UnivariateSpline(r, Qie, k=3, s=0)
    cxcoolint = UnivariateSpline(r, cxcool, k=3, s=0)
    iolPeak = np.where(dE_orb(r) == dE_orb(r).max())

    def f(t, flux, cxcool, Qie, Q_i_nbi, Qnbi_loss, dEdr, iolFlag, peak):
        S = Q_i_nbi(t) - cxcool(t) + Qie(t)
        # Physically, if the IOL peak has occured, everything radially outward should be dFdr = 0.0 since F(r)
        # should equal 0.5 until r=1.0.66
        dEdrval = dEdr(t)
        if t >= peak:
            dEdrval = 0.0
        if iolFlag:
            return S - Qnbi_loss(t) - (flux * (dEdrval + 1 ) / (t + 0.003))
        else:
            return S - (flux * (dEdr(t) + 1 ) / (t + 0.003))

    from scipy.integrate import ode

    flux = ode(f).set_integrator('vode', with_jacobian=False)
    flux.set_initial_value(0., 0.).set_f_params(cxcoolint, Qie_int, en_src_nbi_totint, en_src_nbi_lostint, dE_orb,
                                                iol_adjusted, r[iolPeak])
    dt = a / len(r)
    x, y = [], []
    while flux.successful() and flux.t < a:
        x.append(flux.t + dt)
        y.append(flux.integrate(flux.t + dt))
    flux = UnivariateSpline(x, y, k=3, s=0)

    if verbose:
        plot = plt.figure()
        fig1 = plot.add_subplot(311)
        fig2 = plot.add_subplot(312)
        fig3 = plot.add_subplot(313)

        fig1.scatter(r, Qie_int(r), color="green")
        fig1.scatter(r, cxcoolint(r), color="yellow")
        fig1.scatter(r, en_src_nbi_totint(r), color="red")
        fig1.scatter(r, en_src_nbi_lostint(r), color="black")
        fig1.legend([r"$Q_{ie}$", r"$Q_{cx}$", r"$Q_{nbi,kept}$", r"$Q_{nbi,lost}$"])
        fig1.set_xlim(0.85 * a, a)

        fig2.scatter(r, dE_orb(r), color="red")
        fig2.set_xlim(0.85 * a, a)

        fig3.scatter(r, flux(r))
        fig3.set_xlim(0.85 * a, a)
        plt.show()

    return flux(r)

def calc_Qi_int_method(r, n, T, qie, en_src_nbi_i_kept, cool_rate, iol_adjusted=False, E_orb=None):  # formerly qheat

    Qi = np.zeros(r.shape)

    # Boundary condition at the magnetic axis.
    # Only has the second term, since it's the center value. Also uses delta_r of the next point.
    # If not adjusted for IOL, en_src_nbi_kept = en_src_nbi_tot, so no need for an IOL check.
    Qi[0] = (en_src_nbi_i_kept[0] - cool_rate[0] + qie[0]) * (r[1] - r[0])
    Qi[1] = Qi[0] + (en_src_nbi_i_kept[1] - cool_rate[1] + qie[1]) * (r[1] - r[0])

    # Integral cylindrical form of the energy balance equation.
    # Identical in form to the particle continuity equation, but different source term.
    for n in range(2, len(r)):
        if iol_adjusted:
            # Imported the exp() function for the thermal IOL attenuation.
            Qi[n] = (r[n - 1] / r[n]) * Qi[n - 1] * exp(-(E_orb[n] - E_orb[n - 1])) + (
                        en_src_nbi_i_kept[n] - cool_rate[n] + qie[n]) * (r[n] - r[n - 1])
        else:
            Qi[n] = (r[n - 1] / r[n]) * Qi[n - 1] + (en_src_nbi_i_kept[n] - cool_rate[n] - qie[n]) * (r[n] - r[n - 1])

    return Qi


def calc_qie(n, T, ion_species='D'):
    """Calculates collisional energy transfer between an ion species and electrons
    Reference Equation 4.90 in Stacey's Fusion Plasma Physics Book

    :param n:
    :param T:
    :param ion_species:
    :return:
    """

    if ion_species == 'C':
        zi = 1
        Ti = T.i.J  # assumed carbon is at the same temperature as main ion species
        mi = m_c
    else:
        zi = 1
        Ti = T.i.J
        mi = m_d

    coul_log = calc_coul_log(zi, 1, T.i.J, n.i)  # TODO: check these inputs

    qie = n.e * (zi * e**2)**2 * m_e * coul_log * (1 - Ti / T.e.J) / \
          (2*pi * eps_0**2 * np.sqrt(2*pi*m_e*T.e.J) * mi * (1 + 4*sqrt(pi)/3 * (3*m_e*Ti/(2*mi*T.e.J))**1.5))

    return qie * n.i #multiply this shit by density



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 11 10:21:23 2018

@author: max
"""
from __future__ import division
import numpy as np
from GT3.ImpRadiation import ImpRad
from collections import namedtuple
import sys


def calc_z_0(n):
    z_0 = n.C * 6.0 ** 2 / n.i
    return z_0


def calc_z_eff(n):
    z_eff = (n.i * 1.0 ** 2 + n.C * 6.0 ** 2) / n.e
    return z_eff


# TODO: Generalize this to non carbon impurities. Refer to equations 14.39 in Stacey's book.
def calc_Ci2(n):
    z_0 = calc_z_0(n)
    Ci2 = 1.56 * (1.0 + np.sqrt(2) * z_0) * (1.0 + 0.52 * z_0) / \
          ((1 + 2.65 * z_0) * (1 + 0.285 * z_0) * (z_0 + np.sqrt(0.5 * (1 + (1 / 6)))))
    return Ci2


def calc_fz(n):
    fz = n.C / n.i  # ne?
    return fz


def calc_f0(n):
    f0 = (n.n.s + n.n.t) / n.i  # ne?
    return f0


def calc_f0c(n):
    f0c = n.n.s / n.i  # ne?
    return f0c


def calc_Ce2(n):
    z_eff = calc_z_eff(n)
    Ce2 = 1.5 * (1 - 0.6934 / 1.3167 ** z_eff)
    return Ce2


def calc_C2(n):
    z_0 = calc_z_0(n)

    Ce2 = calc_Ce2(n)
    Ci2 = calc_Ci2(n)
    C2 = Ce2 - z_0 * Ci2
    return C2


def calc_E_ion():
    return 15.466 * 1.6021E-19  # ionization energy of deuterium in Joules


def calc_Lz(n, T):

    # calculate a weighted average neutral temperature
    Tn = (n.n.s*T.n.s + n.n.t*T.n.t) / (n.n.s + n.n.t)

    # calculate a total neutral fraction
    nf = 1.0E-7  #(n.n.s + n.n.t) / n.e

    # obtain interpolations from ImpRad
    imp_C = ImpRad(z=6)
    Lz_interp_C = imp_C.Lz
    dLzdT_interp_C = imp_C.dLzdT

    # obtain values from the interpolations
    Lz_C = Lz_interp_C(np.log10(Tn), np.log10(nf), np.log10(T.e.kev))
    dLzdT_C = dLzdT_interp_C(np.log10(Tn), np.log10(nf), np.log10(T.e.kev))

    return Lz_C, dLzdT_C


def calc_n_marfe(n, sv, T, L, Lz, chi_r):
    """
    Calculates the density limit for MARFE onset

    :param Lz:
    :param n: class instance or namedtuple with density attributes
    :param sv: class instance or namedtuple with cross section attributes
    :param T: class instance or namedtuple with temperature attributes
    :param L: class instance or namedtuple with gradient scale length attributes
    :param chi_r: float or array of floats for chi_radial
    :return: marfe density limit as float or array, depending on the shape of the inputs.
    """
    C2 = calc_C2(n)
    E_ion = calc_E_ion()
    Lz_t = Lz.t
    dLzdT = Lz.ddT

    #Lz_t, dLzdT = calc_Lz_gtedge(T.i.kev*1E3)
    #Lz_t = Lz_t * 1E-13  # now in Jm^3/s
    #dLzdT = dLzdT * 1E-13  # now in Jm^3/s

    nu = 5 / 2
    fz = calc_fz(n)
    f0 = calc_f0(n)
    f0c = calc_f0c(n)

    if False:
        print 'MARFE Index Parameters'
        print ' chi_r = ', chi_r
        print ' nu = ', nu
        print ' L.T^-1 = ', 1/L.T
        print ' L.n^-1 = ', 1/L.n
        print ' C2 = ', C2
        print ' fz = ', fz
        print ' f0 = ', f0
        print ' f0c = ', f0c
        print ' E_ion = ', E_ion
        print ' sv.ion = ', sv.ion
        print ' sv.ion_ddT = ', sv.ion_ddT
        print ' sv.cx = ', sv.cx
        print ' sv.cx_ddT = ', sv.cx_ddT
        print ' sv.el = ', sv.el
        print ' sv.el_ddT = ', sv.el_ddT
        print ' Lz_t = ', Lz_t
        print ' dLzdT = ', dLzdT
        print ' T.i.J = ',T.i.J
        print ' n.e = ', n.e
        print

    term1 = chi_r * (nu * L.T**-2 + (C2 - 1.0) * L.T**-1 * L.n**-1)
    term2 = fz * ((nu + 1 - C2) * Lz_t / (T.i.J+T.e.J)/2 - dLzdT)
    term3 = f0 * (E_ion * sv.ion / T.i.J * (nu - (T.i.J+T.e.J)/2 / sv.ion * sv.ion_ddT))
    term4 = f0c * (3.0 / 2.0 * (sv.cx + sv.el) * (nu - 1.0 - (T.i.J+T.e.J)/2 * (sv.cx_ddT + sv.el_ddT) / (sv.cx + sv.el)))

    if False:
        print
        print 'term1 = ', term1
        print 'term2 = ', term2
        print 'term3 = ', term3
        print 'term4 = ', term4
        print 'term 2+3+4 = ', term2 + term3 + term4

    n_marfe = term1 / \
              (
                      term2 +
                      term3 +
                      term4
              )
    return n_marfe, n.e / n_marfe


class Marfe:

    def __init__(self, inputs=None, core=None):
        """

        :param inputs:
        :param core:
        """
        if core is not None:
            if inputs is not None:
                print 'ignoring \'inputs\' and attempting to use core instance'
            # use 'core' instance to create inputs for calc_n_marfe

            nn_mult = 1.0

            n_xpt = namedtuple('n', 'n e i C W Be')(
                namedtuple('nn', 's t tot')(
                    core.n.n.s[core.xpt_loc][0] * nn_mult,
                    core.n.n.t[core.xpt_loc][0] * nn_mult,
                    core.n.n.tot[core.xpt_loc][0] * nn_mult
                ),
                core.n.e[core.xpt_loc][0],
                core.n.i[core.xpt_loc][0],
                core.n.C[core.xpt_loc][0],
                core.n.C[core.xpt_loc][0],
                core.n.C[core.xpt_loc][0],
            )

            n_obmp = namedtuple('n', 'n e i C')(
                namedtuple('nn', 's t tot')(
                    core.n.n.s[core.obmp_loc][0] * nn_mult,
                    core.n.n.t[core.obmp_loc][0] * nn_mult,
                    core.n.n.tot[core.obmp_loc][0] * nn_mult,
                ),
                core.n.e[core.obmp_loc][0],
                core.n.i[core.obmp_loc][0],
                core.n.C[core.obmp_loc][0]
            )

            n_av = namedtuple('n', 'n e i C')(
                namedtuple('nn', 's t')(
                    (n_xpt.n.s + n_obmp.n.s) / 2,
                    (n_xpt.n.t + n_obmp.n.t) / 2
                ),
                (n_xpt.e + n_obmp.e) / 2,
                (n_xpt.i + n_obmp.i) / 2,
                (n_xpt.C + n_obmp.C) / 2
            )

            T_xpt = namedtuple('T', 'e i n')(
                namedtuple('Te', 'kev ev J')(
                    core.T.e.kev[core.xpt_loc][0],
                    core.T.e.ev[core.xpt_loc][0],
                    core.T.e.J[core.xpt_loc][0]
                ),
                namedtuple('Ti', 'kev ev J')(
                    core.T.i.kev[core.xpt_loc][0],
                    core.T.i.ev[core.xpt_loc][0],
                    core.T.i.J[core.xpt_loc][0]
                ),
                namedtuple('Tn', 's t')(
                    core.T.n.s.kev[core.xpt_loc][0],
                    core.T.n.t.kev[core.xpt_loc][0]
                )
            )

            T_obmp = namedtuple('T', 'e i n')(
                namedtuple('Te', 'kev ev J')(
                    core.T.e.kev[core.obmp_loc][0],
                    core.T.e.ev[core.obmp_loc][0],
                    core.T.e.J[core.obmp_loc][0]
                ),
                namedtuple('Ti', 'kev ev J')(
                    core.T.i.kev[core.obmp_loc][0],
                    core.T.i.ev[core.obmp_loc][0],
                    core.T.i.J[core.obmp_loc][0]
                ),
                namedtuple('Tn', 's t')(
                    core.T.n.s.kev[core.obmp_loc][0],
                    core.T.n.t.kev[core.obmp_loc][0]
                )
            )

            sv_xpt = namedtuple('sv', 'ion ion_ddT el el_ddT cx cx_ddT')(
                core.sv.ion.st[core.xpt_loc][0],
                core.sv.ion.d_dT.st[core.xpt_loc][0],
                core.sv.el.st[core.xpt_loc][0],
                core.sv.el.d_dT.st[core.xpt_loc][0],
                core.sv.cx.st[core.xpt_loc][0],
                core.sv.cx.d_dT.st[core.xpt_loc][0]
            )

            L_xpt = namedtuple('L', 'n T')(
                core.L.n.i[core.xpt_loc][0],
                core.L.T.i[core.xpt_loc][0]
            )

            # L_xpt = namedtuple('L', 'n T')(
            #     1.0/25.0,
            #     1.0/38.0
            # )

            # L_xpt = namedtuple('L', 'n T')(
            #     1.0/15.6,
            #     1.0/14.2
            # )

            L_obmp = namedtuple('L', 'n T')(
                core.L.n.i[core.obmp_loc][0],
                core.L.T.i[core.obmp_loc][0]
            )

            L_ibmp = namedtuple('L', 'n T')(
                core.L.n.i[core.ibmp_loc][0],
                core.L.T.i[core.ibmp_loc][0]
            )

            L_top = namedtuple('L', 'n T')(
                core.L.n.i[core.top_loc][0],
                core.L.T.i[core.top_loc][0]
            )

            L_av = namedtuple('L', 'n T')(
                (core.L.n.i[core.top_loc][0] + core.L.n.i[core.xpt_loc][0] + core.L.n.i[core.ibmp_loc][0] + core.L.n.i[core.obmp_loc][0])/4.0,
                (core.L.T.i[core.top_loc][0] + core.L.T.i[core.xpt_loc][0] + core.L.T.i[core.ibmp_loc][0] + core.L.T.i[core.obmp_loc][0])/4.0
            )

            xpt_weight = 1
            L_av2 = namedtuple('L', 'n T')(
                (xpt_weight * core.L.n.i[core.xpt_loc][0] + (1-xpt_weight) * core.L.n.i[core.obmp_loc][0])/2.0,
                (xpt_weight * core.L.T.i[core.xpt_loc][0] + (1-xpt_weight) * core.L.T.i[core.obmp_loc][0])/2.0
            )

            if False:
                print 'L_xpt.n^-1 = ',1/L_xpt.n
                print 'L_xpt.T^-1 = ',1/L_xpt.T
                print 'L_obmp.n^-1 = ',1/L_obmp.n
                print 'L_obmp.T^-1 = ',1/L_obmp.T
                print 'L_top.n^-1 = ',1/L_top.n
                print 'L_top.T^-1 = ',1/L_top.T
                print 'L_ibmp.n^-1 = ',1/L_ibmp.n
                print 'L_ibmp.T^-1 = ',1/L_ibmp.T
                # sys.exit()

            Lz_xpt = namedtuple('Lz', 't ddT')(
                core.Lz.t[core.xpt_loc][0],
                core.Lz.ddT.t[core.xpt_loc][0]
            )

            chi_r_xpt = core.chi.bohm[core.xpt_loc][0]
            chi_r_top = core.chi.bohm[core.top_loc][0]
            chi_r_ibmp = core.chi.bohm[core.ibmp_loc][0]
            chi_r_obmp = core.chi.bohm[core.obmp_loc][0]
            chi_r_av = (core.chi.bohm[core.xpt_loc][0] + core.chi.bohm[core.top_loc][0] + core.chi.bohm[core.ibmp_loc][0] + core.chi.bohm[core.obmp_loc][0])/4.0

            #self.n_marfe, self.MI = calc_n_marfe(n_xpt, sv_xpt, T_xpt, L_xpt, Lz_xpt, chi_r_xpt)
            self.n_marfe, self.MI = calc_n_marfe(n_xpt, sv_xpt, T_xpt, L_av2, Lz_xpt, chi_r_xpt)

        elif inputs is not None:
            # use 'inputs' to create inputs for calc_n_marfe
            n = inputs.n
            sv = inputs.sv
            T = inputs.T
            L = inputs.L
            Lz = inputs.Lz
            chi_r = inputs.chi_r
            self.n_marfe, self.MI = calc_n_marfe(n, sv, T, L, Lz, chi_r)
        else:
            raise Exception("No inputs provided")

        print
        print '######################################'
        print
        print 'ni = ', n_xpt.i
        print 'ne = ', n_xpt.e
        print 'nn_xpt = ', n_xpt.n.tot
        print 'nn_obmp = ', n_obmp.n.tot
        print 'n_marfe = ',self.n_marfe
        print
        print 'Ti(ev) = ', T_xpt.i.ev
        print 'Te(ev) = ', T_xpt.e.ev
        print
        print 'L_n^-1 (xpt) = ',1/L_xpt.n
        print 'L_T^-1 (xpt) = ',1/L_xpt.T
        print
        print 'L_n^-1 (obmp) = ',1/L_obmp.n
        print 'L_T^-1 (obmp) = ',1/L_obmp.T
        print
        print 'L_n^-1 (av) = ',1/L_av.n
        print 'L_T^-1 (av) = ',1/L_av.T
        print
        print 'nC / ne = ', n_xpt.W / n_xpt.e
        print 'nn / ne (xpt) = ', (n_xpt.n.s + n_xpt.n.t) / n_xpt.e
        print 'nn / ne (obmp) = ', (n_obmp.n.s + n_obmp.n.t) / n_obmp.e
        print 'MI = ', self.MI
        print
        print '######################################'
        print



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import UnivariateSpline

def calc_return_cur(r, part_src_nbi_lost, gamma, ch_d, iol_adjusted=False, F_orb=None):
    # Piper Changes: This calculates the return current from fast and thermal IOL.
    # This is a new quantity that i want to calculate. Big part of my thesis.

    if iol_adjusted:
        diff_F_orb = UnivariateSpline(r, F_orb, k=1, s=0).derivative()(
            r)  # Need the derivative of F_orb with respect to r.
    else:
        diff_F_orb = np.zeros(r.shape)

    Jr_iol = np.zeros(r.shape)
    Jr_iol[0] = (part_src_nbi_lost[0] + (gamma[0] * diff_F_orb[0])) * (r[1] - r[0]) * ch_d
    Jr_iol[1] = Jr_iol[0] + (part_src_nbi_lost[1] + (gamma[1] * diff_F_orb[1])) * (r[1] - r[0]) * ch_d

    # Integral cylindrical form of the continuity of (current?) equation.
    for n in range(2, len(r)):
        Jr_iol[n] = (r[n - 1] / r[n]) * Jr_iol[n - 1] + (part_src_nbi_lost[n] + (gamma[n] * diff_F_orb[n])) * (
                    r[n] - r[n - 1]) * ch_d  # No factor of 2, because only 1 direction.

    return Jr_iol



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from GT3.RadialTransport.Functions.CalcReducedMass import calc_reduced_mass
from GT3.RadialTransport.Functions.CalcCoulLog import calc_coul_log_j_k



def calc_nu_drag(n_j, m_j, v_tor_j, v_tor_k, mbal_rhs, nu_c):
    nu_drag = (mbal_rhs + (n_j * m_j * nu_c * v_tor_k)) / (v_tor_j * n_j * m_j) - nu_c  # Piper Changes: Multiplied V_tor_k in the numerator by n_j*m_j. The algebra was wrong.
    return nu_drag

def calc_nu_j_k(m_j, m_k, z_j, z_k, T_j, n_k):
    m_r = calc_reduced_mass(m_j, m_k)
    coul_log_j_k = calc_coul_log_j_k(z_j, z_k, T_j, n_k)
    C1 = 1/((((4.8E-10)/(1.6E-12))**1.5)*((4.8E-10)**2.5))
    nu_j_k = 3.34*(coul_log_j_k * (z_j**2) * (z_k**2) * 1E-6 * n_k) / \
            (C1*np.sqrt(m_r*1E3)*(T_j**1.5))
    # nu_j_k = (1.2734E14) * (z_j**2) * (z_k**2) * n_k * coul_log_j_k / \
    #        (np.sqrt(m_r * 1E3) * (T_j**1.5))
    return nu_j_k

def calc_nustar(nu90, q, R0_a, vpol):
    nustar = nu90 * abs(q) * R0_a / np.abs(vpol)
    return nustar


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from GT3.RadialTransport.Functions.CalcT90 import calc_t90
from GT3.RadialTransport.Functions.CalcMbalRHS import calc_mbal_rhs
from scipy import constants

m_d = constants.physical_constants['deuteron mass'][0]
m_c = 12 / constants.N_A / 1E3  # in kg

def calc_vtor_d_pert(vtor_C_total, vtor_C_intrin, vtor_D_intrin, mom_src_ext, z, n, T, B_p, gamma_D, gamma_C):
    """
    """
    # Piper Changes: Changed n to n.i and n.C.
    # Changed z for mbal_rhs_C to 6.
    mbal_rhs_D = calc_mbal_rhs(mom_src_ext, z, n.i, B_p, gamma_D)
    mbal_rhs_C = calc_mbal_rhs(mom_src_ext, z, n.C, B_p, gamma_C) # TODO: This z should be zbar.

    nu_c_DC = 1 / calc_t90(m_d, m_c, 1, 6, n.C, T.i.J)

    #vtor_C_total = vtor_fluid_C + vtor_intrin_C
    del_v0 = vtor_D_intrin - vtor_C_intrin

    nu_drag_approx = (mbal_rhs_D + mbal_rhs_C) / ((n.i * m_d + n.C * m_c) * vtor_C_total + n.i * m_d * del_v0)

    del_v1 = (mbal_rhs_D - n.i * m_d * nu_drag_approx * vtor_C_total) / (n.i * m_d * (nu_c_DC + nu_drag_approx))

    vtorDPert = vtor_C_total + del_v1
    return vtorDPert



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from GT3.IOL.Functions.CalcVSep import calc_vsep

def calc_iol_mono_en(z, m, param, thetapts, v_mono, coslist, numcos):
    """calculates IOL for a monoenergetic species with isotropic launch angle (i.e. fast alphas)"""

    v_sep, v_sep_min = calc_vsep(z, m, param)

    # Create the launch angle matrix (used in the calculation of M_orb)
    zeta_matrix = np.zeros(v_sep_min.shape)
    for indx, column in enumerate(zeta_matrix.T):
        zeta_matrix[:, indx] = coslist[indx]

    # F_orb calculation
    #integrand_f = np.where(v_sep_min <= v_mono, 1.0, 0)
    integrand_f = np.heaviside(v_mono - v_sep_min, 1)
    F_orb_1D = np.sum(integrand_f, axis=1) * (2 / numcos) / 2
    F_orb_1D = np.nan_to_num(F_orb_1D)
    F_orb = np.repeat(F_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # M_orb calculation
    #integrand_m = zeta_matrix * np.where(v_sep_min <= v_mono, 1, 0)
    integrand_m = zeta_matrix * np.heaviside(v_mono - v_sep_min, 1)
    M_orb_1D = np.sum(integrand_m, axis=1) * (2 / numcos) / 2
    M_orb_1D = np.nan_to_num(M_orb_1D)
    M_orb = np.repeat(M_orb_1D.reshape(-1, 1), thetapts, axis=1)

    # E_orb calculation
    # E_orb is mathematically identical to F_orb for monoenergetic, isotropic launch angle species
    E_orb = F_orb

    return F_orb, M_orb, E_orb



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from GT3.BeamDeposition.Functions.CalcPwrFrac import calc_pwr_frac
from scipy.interpolate import interp1d
import numpy as np
import os

def prep_nbi_infile(inp, core, index = False, indBeam = False):
    i = index
    if indBeam: pwr_frac = calc_pwr_frac(indBeam['ebeam'])
    else: pwr_frac = calc_pwr_frac(inp.ebeam)

    # pwr_frac = [0.7, 0.2, 0.1]

    # f1=open(inp.nbeams_loc+"inbeams.dat", "w")
    if __name__=="__main__":
        f = open("inputs/inbeams_test.dat", "w")
    else:
        if indBeam: f=open(os.path.dirname(__file__) + "/../inbeams_%s.dat" % str(i), "w")
        else: f = open(os.path.dirname(__file__) + "/../inbeams.dat", "w")
    f.write("$nbin\n")
    f.write("nbeams = 1\n")
    f.write("inbfus = 1\n")
    f.write("amb = 2.0\n")
    f.write("zbeam = 1.0\n")

    if indBeam: f.write("ebeam = " + str(indBeam["ebeam"]) + "\n")
    else: f.write("ebeam = " + str(inp.ebeam) + "\n")

    if indBeam: f.write("pbeam = " + str(indBeam["pbeam"]) + "\n")
    else: f.write("pbeam = " + str(inp.pbeam) + "\n")

    if indBeam: f.write("pbeam = " + str(indBeam["rtang"]) + "\n")
    else: f.write("rtang = " + str(inp.rtang) + "\n")

    f.write("nbshape = 1\n")
    f.write("bwidth = 0.12\n")  # Is this default?
    f.write("bheigh = 0.48\n")
    f.write("bgaussR = 0.066\n")
    f.write("bgaussZ = 0.18\n")
    f.write("bzpos = 0.0\n")
    f.write("nbptype = 1\n")
    f.write("maxiter = 2\n")
    f.write("pwrfrac(1, 1) = " + str(pwr_frac[0]) + "   " + str(pwr_frac[1]) + "   " + str(pwr_frac[2]) + "\n")
    f.write("a = " + str(core.a) + "\n")
    f.write("r0 = " + str(core.R0_a) + "\n")
    f.write("b0 = " + str(inp.BT0) + "\n")
    f.write("n = 51\n")
    f.write("e0 = " + str(core.kappa_vals.axis) + "\n")
    f.write("ea = " + str(core.kappa_vals.sep) + "\n")
    f.write("shft0 = " + str(core.shaf_shift) + "\n")
    f.write("nion = 2\n")
    f.write("aion = 2.0 12.0\n")
    f.write("zion = 1.0 6.0\n")

    rho_nbi = np.linspace(0, 1, 51)

    ni_nbi = interp1d(core.rho[:, 0], core.n.i[:, 0])(rho_nbi)
    ne_nbi = interp1d(core.rho[:, 0], core.n.e[:, 0])(rho_nbi)
    Ti_nbi = interp1d(core.rho[:, 0], core.T.i.kev[:, 0])(rho_nbi)
    Te_nbi = interp1d(core.rho[:, 0], core.T.e.kev[:, 0])(rho_nbi)

    for i, v in enumerate(rho_nbi):
        f.write('ni20(' + str(i + 1) + ', 1) = ' + str(ni_nbi[i] * 1E-20) + '\n')
    for i, v in enumerate(rho_nbi):
        f.write('ne20(' + str(i + 1) + ') = ' + str(ne_nbi[i] * 1E-20) + '\n')
    for i, v in enumerate(rho_nbi):
        f.write('tikev(' + str(i + 1) + ') = ' + str(Ti_nbi[i]) + '\n')
    for i, v in enumerate(rho_nbi):
        f.write('tekev(' + str(i + 1) + ') = ' + str(Te_nbi[i]) + '\n')
    f.write("$end\n")
    f.close()


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import interp2d, UnivariateSpline, griddata
from scipy import constants

e = constants.elementary_charge

def calc_svcx_st(T):
    tint = np.array([-1, 0, 1, 2, 3])
    tnnt = np.array([0, 1, 2])

    cx = np.array([[-1.4097E+01, -1.3921E+01, -1.3553E+01, -1.4097E+01, -1.3921E+01],
                   [-1.3553E+01, -1.3921E+01, -1.3824E+01, -1.3538E+01, -1.3553E+01],
                   [-1.3538E+01, -1.3432E+01, -1.3553E+01, -1.3538E+01, -1.3432E+01]])

    interp1 = interp2d(tint, tnnt, cx)

    Ti_exps = np.linspace(-1, 3, 100)
    Tn_exps = np.linspace(0, 2, 100)
    svcx_vals = 10.0 ** (interp1(Ti_exps, Tn_exps))  # in m^3/s

    Ti_vals = np.logspace(-1, 3, 100) * e  # in joules
    Tn_vals = np.logspace(0, 2, 100) * e  # in joules

    dsvcx_dTi_vals = np.gradient(svcx_vals, Ti_vals, axis=0)

    Ti_vals2d, Tn_vals2d = np.meshgrid(Ti_vals, Tn_vals)

    Ti_mod = np.where(T.i.ev > 1E3, 1E3 * e, T.i.ev * e)
    Tn_mod = np.zeros(Ti_mod.shape) + 2.0 * e

    sv_cx = griddata(np.column_stack((Ti_vals2d.flatten(), Tn_vals2d.flatten())),
                          svcx_vals.flatten(),
                          (Ti_mod, Tn_mod),
                          method='linear', rescale=False)

    dsv_cx_dT = griddata(np.column_stack((Ti_vals2d.flatten(), Tn_vals2d.flatten())),
                              dsvcx_dTi_vals.flatten(),
                              (Ti_mod, Tn_mod),
                              method='linear', rescale=False)

    return sv_cx, dsv_cx_dT

def calc_svel_st(T):
    tint = np.array([-1, 0, 1, 2, 3])
    tnnt = np.array([0, 1, 2])

    elast = np.array([[-1.3569E+01, -1.3337E+01, -1.3036E+01, -1.3569E+01, -1.3337E+01],
                      [-1.3036E+01, -1.3337E+01, -1.3167E+01, -1.3046E+01, -1.3036E+01],
                      [-1.3046E+01, -1.2796E+01, -1.3036E+01, -1.3046E+01, -1.2796E+01]])

    interp1 = interp2d(tint, tnnt, elast)

    Ti_exps = np.linspace(-1, 3, 100)
    Tn_exps = np.linspace(0, 2, 100)
    svel_vals = 10.0 ** (interp1(Ti_exps, Tn_exps))  # in m^3/s

    Ti_vals = np.logspace(-1, 3, 100) * e  # in joules
    Tn_vals = np.logspace(0, 2, 100) * e  # in joules

    dsvel_dTi_vals = np.gradient(svel_vals, Ti_vals, axis=0)

    Ti_vals2d, Tn_vals2d = np.meshgrid(Ti_vals, Tn_vals)

    Ti_mod = np.where(T.i.ev > 1E3, 1E3 * e, T.i.ev * e)
    Tn_mod = np.zeros(Ti_mod.shape) + 2.0 * e

    sv_el = griddata(np.column_stack((Ti_vals2d.flatten(), Tn_vals2d.flatten())),
                          svel_vals.flatten(),
                          (Ti_mod, Tn_mod),
                          method='linear', rescale=False)
    dsv_el_dT = griddata(np.column_stack((Ti_vals2d.flatten(), Tn_vals2d.flatten())),
                              dsvel_dTi_vals.flatten(),
                              (Ti_mod, Tn_mod),
                              method='linear', rescale=False)

    return sv_el, dsv_el_dT

def calc_svfus(T, mode='dd'):
    def sigv(T, mode):  # function takes T in kev
        if mode == 'dt':
            B_G = 34.3827
            m_rc2 = 1124656

            C1 = 1.17302E-9
            C2 = 1.51361E-2
            C3 = 7.51886E-2
            C4 = 4.60643E-3
            C5 = 1.35000E-2
            C6 = -1.06750E-4
            C7 = 1.36600E-5

            theta = T / (1.0 - (T * (C2 + T * (C4 + T * C6))) / (1.0 + T * (C3 + T * (C5 + T * C7))))
            xi = (B_G ** 2.0 / (4.0 * theta)) ** (1.0 / 3.0)
            sigv = C1 * theta * np.sqrt(xi / (m_rc2 * T ** 3.0)) * np.exp(-3.0 * xi)
            sigv = sigv / 1.0E6  # convert from cm^3/s to m^3/s

        elif mode == 'dd':

            B_G = 31.3970
            m_rc2 = 937814

            # first for the D(d, p)T reaction
            C1_1 = 5.65718E-12
            C2_1 = 3.41267E-3
            C3_1 = 1.99167E-3
            C4_1 = 0.0
            C5_1 = 1.05060E-5
            C6_1 = 0.0
            C7_1 = 0.0

            theta_1 = T / (
                        1.0 - (T * (C2_1 + T * (C4_1 + T * C6_1))) / (1.0 + T * (C3_1 + T * (C5_1 + T * C7_1))))
            xi_1 = (B_G ** 2.0 / (4.0 * theta_1)) ** (1.0 / 3.0)
            sigv_1 = C1_1 * theta_1 * np.sqrt(xi_1 / (m_rc2 * T ** 3.0)) * np.exp(-3.0 * xi_1)

            # then for the D(d, n)He3 reaction

            C1_2 = 5.43360E-12
            C2_2 = 5.85778E-3
            C3_2 = 7.68222E-3
            C4_2 = 0.0
            C5_2 = -2.96400E-6
            C6_2 = 0.0
            C7_2 = 0.0

            theta_2 = T / (
                        1.0 - (T * (C2_2 + T * (C4_2 + T * C6_2))) / (1.0 + T * (C3_2 + T * (C5_2 + T * C7_2))))
            xi_2 = (B_G ** 2.0 / (4.0 * theta_2)) ** (1.0 / 3.0)
            sigv_2 = C1_2 * theta_2 * np.sqrt(xi_2 / (m_rc2 * T ** 3.0)) * np.exp(-3.0 * xi_2)

            sigv = (0.5 * sigv_1 + 0.5 * sigv_2) / 1.0E6  # convert from cm^3/s to m^3/s
        elif mode == 'dHe3':
            B_G = 68.7508
            m_rc2 = 1124572

            C1 = 5.51036E-10
            C2 = 6.41918E-3
            C3 = -2.02896E-3
            C4 = -1.91080E-5
            C5 = 1.35776E-4
            C6 = 0
            C7 = 0

            theta = T / (1.0 - (T * (C2 + T * (C4 + T * C6))) / (1.0 + T * (C3 + T * (C5 + T * C7))))
            xi = (B_G ** 2.0 / (4.0 * theta)) ** (1.0 / 3.0)
            sigv = C1 * theta * np.sqrt(xi / (m_rc2 * T ** 3.0)) * np.exp(-3.0 * xi)
            sigv = sigv / 1.0E6  # convert from cm^3/s to m^3/s
        return sigv

    # create logspace over the relevant temperature range
    # (bosch hale technically only valid over 0.2 - 100 kev)
    Ti_range = np.logspace(-1, 2, 1000)  # values in kev
    sigv_fus_range = sigv(Ti_range, mode=mode)  # in m^3/s
    sigv_fus_interp = UnivariateSpline(Ti_range * 1.0E3 * e, sigv_fus_range, s=0)  # converted to Joules
    sv_fus = sigv_fus_interp(T.i.J)
    dsv_fus_dT = sigv_fus_interp.derivative()(T.i.J)

    return sv_fus, dsv_fus_dT

def calc_svion_st(T):
    # TODO: configure so it can use any of the cross section libraries
    # currently using the Stacey-Thomas cross sections
    T_exps_fit = np.array([-1, 0, 1, 2, 3, 4, 5])
    sigv_exps_fit = np.array([-2.8523E+01, -1.7745E+01, -1.3620E+01,
                              -1.3097E+01, -1.3301E+01, -1.3301E+01, -1.3301E+01])
    interp1 = UnivariateSpline(T_exps_fit, sigv_exps_fit, s=0)

    T_exps_range = np.linspace(-1, 5, 1000)
    sigv_vals_range = 10.0 ** interp1(T_exps_range)  # in m^3/s

    T_vals_range = np.logspace(-1, 5, 1000) * e  # in joules
    interp2 = UnivariateSpline(T_vals_range, sigv_vals_range, s=0)

    sv_ion = interp2(T.i.J)
    dsv_ion_dT = interp2.derivative()(T.i.J)

    return sv_ion, dsv_ion_dT

def calc_svrec_st(n, T):
    # # TODO: check this calculation. -MH
    # znint = np.array([16, 18, 20, 21, 22])
    # Tint = np.array([-1, 0, 1, 2, 3])
    #
    # rec = np.array([[-1.7523E+01, -1.6745E+01, -1.5155E+01, -1.4222E+01, -1.3301E+01],
    #                 [-1.8409E+01, -1.8398E+01, -1.8398E+01, -1.7886E+01, -1.7000E+01],
    #                 [-1.9398E+01, -1.9398E+01, -1.9398E+01, -1.9398E+01, -1.9398E+01],
    #                 [-2.0155E+01, -2.0155E+01, -2.0155E+01, -2.0155E+01, -2.0155E+01],
    #                 [-2.1000E+01, -2.1000E+01, -2.1000E+01, -2.1000E+01, -2.1000E+01]])
    #
    # interp1 = interp2d(znint, Tint, rec)
    #
    # zni_exps = np.linspace(16, 22, 100)
    # Ti_exps = np.linspace(-1, 3, 100)
    # svrec_vals = 10.0 ** (interp1(zni_exps, Ti_exps))  # in m^3/s
    #
    # zni_vals = np.logspace(16, 22, 100)
    # Ti_vals = np.logspace(-1, 3, 100) * e  # in joules
    #
    # dsvrec_dTi_vals = np.gradient(svrec_vals, Ti_vals, axis=0)
    #
    # zni_vals2d, Ti_vals2d = np.meshgrid(zni_vals, Ti_vals)
    #
    # zni_mod = np.where(n.i > 1E22, 1E22, n.i)
    # zni_mod = np.where(n.i < 1E16, 1E16, zni_mod)
    # Ti_mod = np.where(T.i.ev > 1E3, 1E3 * e, T.i.ev * e)
    # Ti_mod = np.where(T.i.ev < 1E-1, 1E-1 * e, Ti_mod)
    #
    # plt.semilogx(zni_vals2d.flatten(), Ti_vals2d.flatten())
    # plt.show()
    # print np.column_stack((zni_vals2d.flatten(), Ti_vals2d.flatten()))
    # sys.exit()
    # sv_rec = griddata(np.column_stack((zni_vals2d.flatten(), Ti_vals2d.flatten())),
    #                   svrec_vals.flatten(),
    #                   (zni_mod, Ti_mod),
    #                   method='linear', rescale=False)
    #
    # dsv_rec_dT = griddata(np.column_stack((zni_vals2d.flatten(), Ti_vals2d.flatten())),
    #                            dsvrec_dTi_vals.flatten(),
    #                            (zni_mod, Ti_mod),
    #                            method='linear', rescale=False)

    # return sv_rec, dsv_rec_dT
    return 0, 0


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from shapely.geometry import Point

def isinline(pt, line):
    pt_s = Point(pt)
    dist = line.distance(pt_s)
    if dist < 1E-6:
        return True
    else:
        return False